# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class OrmRunExecution(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'arguments': 'dict[str, object]',
        'created_dt': 'datetime',
        'edges': 'OrmRunExecutionEdges',
        'execution_created_by': 'int',
        'exposed_services': 'dict[str, object]',
        'final_kernel_resource_spec': 'dict[str, object]',
        'id': 'int',
        'immutable_slug': 'str',
        'interactive_config': 'dict[str, object]',
        'interactive_end_dt': 'datetime',
        'local_experiment_id': 'int',
        'message': 'str',
        'metrics_summary': 'dict[str, object]',
        'organization_id': 'int',
        'pipeline_step_execution_id': 'int',
        'progress_percent': 'int',
        'project_id': 'int',
        'run_spec_id': 'int',
        'star': 'bool',
        'updated_dt': 'datetime'
    }

    attribute_map = {
        'arguments': 'arguments',
        'created_dt': 'created_dt',
        'edges': 'edges',
        'execution_created_by': 'execution_created_by',
        'exposed_services': 'exposed_services',
        'final_kernel_resource_spec': 'final_kernel_resource_spec',
        'id': 'id',
        'immutable_slug': 'immutable_slug',
        'interactive_config': 'interactive_config',
        'interactive_end_dt': 'interactive_end_dt',
        'local_experiment_id': 'local_experiment_id',
        'message': 'message',
        'metrics_summary': 'metrics_summary',
        'organization_id': 'organization_id',
        'pipeline_step_execution_id': 'pipeline_step_execution_id',
        'progress_percent': 'progress_percent',
        'project_id': 'project_id',
        'run_spec_id': 'run_spec_id',
        'star': 'star',
        'updated_dt': 'updated_dt'
    }

    def __init__(self, arguments=None, created_dt=None, edges=None, execution_created_by=None, exposed_services=None, final_kernel_resource_spec=None, id=None, immutable_slug=None, interactive_config=None, interactive_end_dt=None, local_experiment_id=None, message=None, metrics_summary=None, organization_id=None, pipeline_step_execution_id=None, progress_percent=None, project_id=None, run_spec_id=None, star=None, updated_dt=None, local_vars_configuration=None):  # noqa: E501
        """OrmRunExecution - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._arguments = None
        self._created_dt = None
        self._edges = None
        self._execution_created_by = None
        self._exposed_services = None
        self._final_kernel_resource_spec = None
        self._id = None
        self._immutable_slug = None
        self._interactive_config = None
        self._interactive_end_dt = None
        self._local_experiment_id = None
        self._message = None
        self._metrics_summary = None
        self._organization_id = None
        self._pipeline_step_execution_id = None
        self._progress_percent = None
        self._project_id = None
        self._run_spec_id = None
        self._star = None
        self._updated_dt = None
        self.discriminator = None

        if arguments is not None:
            self.arguments = arguments
        if created_dt is not None:
            self.created_dt = created_dt
        if edges is not None:
            self.edges = edges
        self.execution_created_by = execution_created_by
        if exposed_services is not None:
            self.exposed_services = exposed_services
        if final_kernel_resource_spec is not None:
            self.final_kernel_resource_spec = final_kernel_resource_spec
        if id is not None:
            self.id = id
        if immutable_slug is not None:
            self.immutable_slug = immutable_slug
        if interactive_config is not None:
            self.interactive_config = interactive_config
        self.interactive_end_dt = interactive_end_dt
        self.local_experiment_id = local_experiment_id
        if message is not None:
            self.message = message
        if metrics_summary is not None:
            self.metrics_summary = metrics_summary
        if organization_id is not None:
            self.organization_id = organization_id
        self.pipeline_step_execution_id = pipeline_step_execution_id
        self.progress_percent = progress_percent
        self.project_id = project_id
        if run_spec_id is not None:
            self.run_spec_id = run_spec_id
        if star is not None:
            self.star = star
        if updated_dt is not None:
            self.updated_dt = updated_dt

    @property
    def arguments(self):
        """Gets the arguments of this OrmRunExecution.  # noqa: E501


        :return: The arguments of this OrmRunExecution.  # noqa: E501
        :rtype: dict[str, object]
        """
        return self._arguments

    @arguments.setter
    def arguments(self, arguments):
        """Sets the arguments of this OrmRunExecution.


        :param arguments: The arguments of this OrmRunExecution.  # noqa: E501
        :type arguments: dict[str, object]
        """

        self._arguments = arguments

    @property
    def created_dt(self):
        """Gets the created_dt of this OrmRunExecution.  # noqa: E501


        :return: The created_dt of this OrmRunExecution.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this OrmRunExecution.


        :param created_dt: The created_dt of this OrmRunExecution.  # noqa: E501
        :type created_dt: datetime
        """

        self._created_dt = created_dt

    @property
    def edges(self):
        """Gets the edges of this OrmRunExecution.  # noqa: E501


        :return: The edges of this OrmRunExecution.  # noqa: E501
        :rtype: OrmRunExecutionEdges
        """
        return self._edges

    @edges.setter
    def edges(self, edges):
        """Sets the edges of this OrmRunExecution.


        :param edges: The edges of this OrmRunExecution.  # noqa: E501
        :type edges: OrmRunExecutionEdges
        """

        self._edges = edges

    @property
    def execution_created_by(self):
        """Gets the execution_created_by of this OrmRunExecution.  # noqa: E501


        :return: The execution_created_by of this OrmRunExecution.  # noqa: E501
        :rtype: int
        """
        return self._execution_created_by

    @execution_created_by.setter
    def execution_created_by(self, execution_created_by):
        """Sets the execution_created_by of this OrmRunExecution.


        :param execution_created_by: The execution_created_by of this OrmRunExecution.  # noqa: E501
        :type execution_created_by: int
        """

        self._execution_created_by = execution_created_by

    @property
    def exposed_services(self):
        """Gets the exposed_services of this OrmRunExecution.  # noqa: E501


        :return: The exposed_services of this OrmRunExecution.  # noqa: E501
        :rtype: dict[str, object]
        """
        return self._exposed_services

    @exposed_services.setter
    def exposed_services(self, exposed_services):
        """Sets the exposed_services of this OrmRunExecution.


        :param exposed_services: The exposed_services of this OrmRunExecution.  # noqa: E501
        :type exposed_services: dict[str, object]
        """

        self._exposed_services = exposed_services

    @property
    def final_kernel_resource_spec(self):
        """Gets the final_kernel_resource_spec of this OrmRunExecution.  # noqa: E501


        :return: The final_kernel_resource_spec of this OrmRunExecution.  # noqa: E501
        :rtype: dict[str, object]
        """
        return self._final_kernel_resource_spec

    @final_kernel_resource_spec.setter
    def final_kernel_resource_spec(self, final_kernel_resource_spec):
        """Sets the final_kernel_resource_spec of this OrmRunExecution.


        :param final_kernel_resource_spec: The final_kernel_resource_spec of this OrmRunExecution.  # noqa: E501
        :type final_kernel_resource_spec: dict[str, object]
        """

        self._final_kernel_resource_spec = final_kernel_resource_spec

    @property
    def id(self):
        """Gets the id of this OrmRunExecution.  # noqa: E501


        :return: The id of this OrmRunExecution.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this OrmRunExecution.


        :param id: The id of this OrmRunExecution.  # noqa: E501
        :type id: int
        """

        self._id = id

    @property
    def immutable_slug(self):
        """Gets the immutable_slug of this OrmRunExecution.  # noqa: E501


        :return: The immutable_slug of this OrmRunExecution.  # noqa: E501
        :rtype: str
        """
        return self._immutable_slug

    @immutable_slug.setter
    def immutable_slug(self, immutable_slug):
        """Sets the immutable_slug of this OrmRunExecution.


        :param immutable_slug: The immutable_slug of this OrmRunExecution.  # noqa: E501
        :type immutable_slug: str
        """

        self._immutable_slug = immutable_slug

    @property
    def interactive_config(self):
        """Gets the interactive_config of this OrmRunExecution.  # noqa: E501


        :return: The interactive_config of this OrmRunExecution.  # noqa: E501
        :rtype: dict[str, object]
        """
        return self._interactive_config

    @interactive_config.setter
    def interactive_config(self, interactive_config):
        """Sets the interactive_config of this OrmRunExecution.


        :param interactive_config: The interactive_config of this OrmRunExecution.  # noqa: E501
        :type interactive_config: dict[str, object]
        """

        self._interactive_config = interactive_config

    @property
    def interactive_end_dt(self):
        """Gets the interactive_end_dt of this OrmRunExecution.  # noqa: E501


        :return: The interactive_end_dt of this OrmRunExecution.  # noqa: E501
        :rtype: datetime
        """
        return self._interactive_end_dt

    @interactive_end_dt.setter
    def interactive_end_dt(self, interactive_end_dt):
        """Sets the interactive_end_dt of this OrmRunExecution.


        :param interactive_end_dt: The interactive_end_dt of this OrmRunExecution.  # noqa: E501
        :type interactive_end_dt: datetime
        """

        self._interactive_end_dt = interactive_end_dt

    @property
    def local_experiment_id(self):
        """Gets the local_experiment_id of this OrmRunExecution.  # noqa: E501


        :return: The local_experiment_id of this OrmRunExecution.  # noqa: E501
        :rtype: int
        """
        return self._local_experiment_id

    @local_experiment_id.setter
    def local_experiment_id(self, local_experiment_id):
        """Sets the local_experiment_id of this OrmRunExecution.


        :param local_experiment_id: The local_experiment_id of this OrmRunExecution.  # noqa: E501
        :type local_experiment_id: int
        """

        self._local_experiment_id = local_experiment_id

    @property
    def message(self):
        """Gets the message of this OrmRunExecution.  # noqa: E501


        :return: The message of this OrmRunExecution.  # noqa: E501
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """Sets the message of this OrmRunExecution.


        :param message: The message of this OrmRunExecution.  # noqa: E501
        :type message: str
        """

        self._message = message

    @property
    def metrics_summary(self):
        """Gets the metrics_summary of this OrmRunExecution.  # noqa: E501


        :return: The metrics_summary of this OrmRunExecution.  # noqa: E501
        :rtype: dict[str, object]
        """
        return self._metrics_summary

    @metrics_summary.setter
    def metrics_summary(self, metrics_summary):
        """Sets the metrics_summary of this OrmRunExecution.


        :param metrics_summary: The metrics_summary of this OrmRunExecution.  # noqa: E501
        :type metrics_summary: dict[str, object]
        """

        self._metrics_summary = metrics_summary

    @property
    def organization_id(self):
        """Gets the organization_id of this OrmRunExecution.  # noqa: E501


        :return: The organization_id of this OrmRunExecution.  # noqa: E501
        :rtype: int
        """
        return self._organization_id

    @organization_id.setter
    def organization_id(self, organization_id):
        """Sets the organization_id of this OrmRunExecution.


        :param organization_id: The organization_id of this OrmRunExecution.  # noqa: E501
        :type organization_id: int
        """

        self._organization_id = organization_id

    @property
    def pipeline_step_execution_id(self):
        """Gets the pipeline_step_execution_id of this OrmRunExecution.  # noqa: E501


        :return: The pipeline_step_execution_id of this OrmRunExecution.  # noqa: E501
        :rtype: int
        """
        return self._pipeline_step_execution_id

    @pipeline_step_execution_id.setter
    def pipeline_step_execution_id(self, pipeline_step_execution_id):
        """Sets the pipeline_step_execution_id of this OrmRunExecution.


        :param pipeline_step_execution_id: The pipeline_step_execution_id of this OrmRunExecution.  # noqa: E501
        :type pipeline_step_execution_id: int
        """

        self._pipeline_step_execution_id = pipeline_step_execution_id

    @property
    def progress_percent(self):
        """Gets the progress_percent of this OrmRunExecution.  # noqa: E501


        :return: The progress_percent of this OrmRunExecution.  # noqa: E501
        :rtype: int
        """
        return self._progress_percent

    @progress_percent.setter
    def progress_percent(self, progress_percent):
        """Sets the progress_percent of this OrmRunExecution.


        :param progress_percent: The progress_percent of this OrmRunExecution.  # noqa: E501
        :type progress_percent: int
        """

        self._progress_percent = progress_percent

    @property
    def project_id(self):
        """Gets the project_id of this OrmRunExecution.  # noqa: E501


        :return: The project_id of this OrmRunExecution.  # noqa: E501
        :rtype: int
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this OrmRunExecution.


        :param project_id: The project_id of this OrmRunExecution.  # noqa: E501
        :type project_id: int
        """

        self._project_id = project_id

    @property
    def run_spec_id(self):
        """Gets the run_spec_id of this OrmRunExecution.  # noqa: E501


        :return: The run_spec_id of this OrmRunExecution.  # noqa: E501
        :rtype: int
        """
        return self._run_spec_id

    @run_spec_id.setter
    def run_spec_id(self, run_spec_id):
        """Sets the run_spec_id of this OrmRunExecution.


        :param run_spec_id: The run_spec_id of this OrmRunExecution.  # noqa: E501
        :type run_spec_id: int
        """

        self._run_spec_id = run_spec_id

    @property
    def star(self):
        """Gets the star of this OrmRunExecution.  # noqa: E501


        :return: The star of this OrmRunExecution.  # noqa: E501
        :rtype: bool
        """
        return self._star

    @star.setter
    def star(self, star):
        """Sets the star of this OrmRunExecution.


        :param star: The star of this OrmRunExecution.  # noqa: E501
        :type star: bool
        """

        self._star = star

    @property
    def updated_dt(self):
        """Gets the updated_dt of this OrmRunExecution.  # noqa: E501


        :return: The updated_dt of this OrmRunExecution.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this OrmRunExecution.


        :param updated_dt: The updated_dt of this OrmRunExecution.  # noqa: E501
        :type updated_dt: datetime
        """

        self._updated_dt = updated_dt

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmRunExecution):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmRunExecution):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class OrmModelServiceRolloutSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'no_rollback_on_failure': 'bool',
        'rollback_of_rollout_number': 'int',
        'step_timeout_second': 'int',
        'steps': 'list[OrmModelServiceRolloutStepWrapper]'
    }

    attribute_map = {
        'no_rollback_on_failure': 'no_rollback_on_failure',
        'rollback_of_rollout_number': 'rollback_of_rollout_number',
        'step_timeout_second': 'step_timeout_second',
        'steps': 'steps'
    }

    def __init__(self, no_rollback_on_failure=None, rollback_of_rollout_number=None, step_timeout_second=None, steps=None, local_vars_configuration=None):  # noqa: E501
        """OrmModelServiceRolloutSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._no_rollback_on_failure = None
        self._rollback_of_rollout_number = None
        self._step_timeout_second = None
        self._steps = None
        self.discriminator = None

        if no_rollback_on_failure is not None:
            self.no_rollback_on_failure = no_rollback_on_failure
        self.rollback_of_rollout_number = rollback_of_rollout_number
        if step_timeout_second is not None:
            self.step_timeout_second = step_timeout_second
        self.steps = steps

    @property
    def no_rollback_on_failure(self):
        """Gets the no_rollback_on_failure of this OrmModelServiceRolloutSpec.  # noqa: E501


        :return: The no_rollback_on_failure of this OrmModelServiceRolloutSpec.  # noqa: E501
        :rtype: bool
        """
        return self._no_rollback_on_failure

    @no_rollback_on_failure.setter
    def no_rollback_on_failure(self, no_rollback_on_failure):
        """Sets the no_rollback_on_failure of this OrmModelServiceRolloutSpec.


        :param no_rollback_on_failure: The no_rollback_on_failure of this OrmModelServiceRolloutSpec.  # noqa: E501
        :type no_rollback_on_failure: bool
        """

        self._no_rollback_on_failure = no_rollback_on_failure

    @property
    def rollback_of_rollout_number(self):
        """Gets the rollback_of_rollout_number of this OrmModelServiceRolloutSpec.  # noqa: E501


        :return: The rollback_of_rollout_number of this OrmModelServiceRolloutSpec.  # noqa: E501
        :rtype: int
        """
        return self._rollback_of_rollout_number

    @rollback_of_rollout_number.setter
    def rollback_of_rollout_number(self, rollback_of_rollout_number):
        """Sets the rollback_of_rollout_number of this OrmModelServiceRolloutSpec.


        :param rollback_of_rollout_number: The rollback_of_rollout_number of this OrmModelServiceRolloutSpec.  # noqa: E501
        :type rollback_of_rollout_number: int
        """

        self._rollback_of_rollout_number = rollback_of_rollout_number

    @property
    def step_timeout_second(self):
        """Gets the step_timeout_second of this OrmModelServiceRolloutSpec.  # noqa: E501


        :return: The step_timeout_second of this OrmModelServiceRolloutSpec.  # noqa: E501
        :rtype: int
        """
        return self._step_timeout_second

    @step_timeout_second.setter
    def step_timeout_second(self, step_timeout_second):
        """Sets the step_timeout_second of this OrmModelServiceRolloutSpec.


        :param step_timeout_second: The step_timeout_second of this OrmModelServiceRolloutSpec.  # noqa: E501
        :type step_timeout_second: int
        """

        self._step_timeout_second = step_timeout_second

    @property
    def steps(self):
        """Gets the steps of this OrmModelServiceRolloutSpec.  # noqa: E501


        :return: The steps of this OrmModelServiceRolloutSpec.  # noqa: E501
        :rtype: list[OrmModelServiceRolloutStepWrapper]
        """
        return self._steps

    @steps.setter
    def steps(self, steps):
        """Sets the steps of this OrmModelServiceRolloutSpec.


        :param steps: The steps of this OrmModelServiceRolloutSpec.  # noqa: E501
        :type steps: list[OrmModelServiceRolloutStepWrapper]
        """
        if self.local_vars_configuration.client_side_validation and steps is None:  # noqa: E501
            raise ValueError("Invalid value for `steps`, must not be `None`")  # noqa: E501

        self._steps = steps

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmModelServiceRolloutSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmModelServiceRolloutSpec):
            return True

        return self.to_dict() != other.to_dict()



import os
from typing import List

from vessl.openapi_client.models import ResponseSSHKeyInfo
from vessl.openapi_client.models.ssh_key_create_api_input import SSHKeyCreateAPIInput
from vessl import vessl_api


def list_ssh_keys() -> List[ResponseSSHKeyInfo]:
    """List ssh public keys.

    Example:
        ```python
        vessl.list_ssh_keys()
        ```
    """
    return vessl_api.s_sh_key_list_api().ssh_keys


def create_ssh_key(key_path: str, key_name: str, ssh_public_key_value: str) -> ResponseSSHKeyInfo:
    """Create a SSH public key.

    Args:
        key_path(str): SSH public key path.
        key_name(str): SSH public key name,
        ssh_public_key_value(str): SSH public key value.

    Example:
        ```python
        vessl.create_ssh_key(
            key_path="/Users/johndoe/.ssh/id_ed25519.pub",
            key_name="john@abcd.com",
            ssh_public_key_value="ssh-public-key-value",
        )
        ```
    """
    return vessl_api.s_sh_key_create_api(
        ssh_key_create_api_input=SSHKeyCreateAPIInput(
            filename=os.path.basename(key_path),
            name=key_name,
            public_key=ssh_public_key_value,
        )
    )


def delete_ssh_key(key_id: int) -> object:
    """Delete the ssh public key.

    Args:
        key_id(int): Key ID.

    Example:
        ```python
        vessl.delete_ssh_key(
            key_id=123456,
        )
        ```
    """
    return vessl_api.s_sh_key_delete_api(ssh_key_id=key_id)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class OrmWalletHistory(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'additional_data': 'dict[str, object]',
        'artifact_id': 'int',
        'balance_from': 'int',
        'balance_to': 'int',
        'cluster_id': 'int',
        'coupon_redemption_id': 'int',
        'created_dt': 'datetime',
        'edges': 'OrmWalletHistoryEdges',
        'id': 'int',
        'organization_id': 'int',
        'reason': 'str',
        'type': 'str',
        'updated_dt': 'datetime',
        'wallet_id': 'int',
        'workload_id': 'int'
    }

    attribute_map = {
        'additional_data': 'additional_data',
        'artifact_id': 'artifact_id',
        'balance_from': 'balance_from',
        'balance_to': 'balance_to',
        'cluster_id': 'cluster_id',
        'coupon_redemption_id': 'coupon_redemption_id',
        'created_dt': 'created_dt',
        'edges': 'edges',
        'id': 'id',
        'organization_id': 'organization_id',
        'reason': 'reason',
        'type': 'type',
        'updated_dt': 'updated_dt',
        'wallet_id': 'wallet_id',
        'workload_id': 'workload_id'
    }

    def __init__(self, additional_data=None, artifact_id=None, balance_from=None, balance_to=None, cluster_id=None, coupon_redemption_id=None, created_dt=None, edges=None, id=None, organization_id=None, reason=None, type=None, updated_dt=None, wallet_id=None, workload_id=None, local_vars_configuration=None):  # noqa: E501
        """OrmWalletHistory - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._additional_data = None
        self._artifact_id = None
        self._balance_from = None
        self._balance_to = None
        self._cluster_id = None
        self._coupon_redemption_id = None
        self._created_dt = None
        self._edges = None
        self._id = None
        self._organization_id = None
        self._reason = None
        self._type = None
        self._updated_dt = None
        self._wallet_id = None
        self._workload_id = None
        self.discriminator = None

        if additional_data is not None:
            self.additional_data = additional_data
        self.artifact_id = artifact_id
        if balance_from is not None:
            self.balance_from = balance_from
        if balance_to is not None:
            self.balance_to = balance_to
        self.cluster_id = cluster_id
        self.coupon_redemption_id = coupon_redemption_id
        if created_dt is not None:
            self.created_dt = created_dt
        if edges is not None:
            self.edges = edges
        if id is not None:
            self.id = id
        self.organization_id = organization_id
        if reason is not None:
            self.reason = reason
        if type is not None:
            self.type = type
        if updated_dt is not None:
            self.updated_dt = updated_dt
        if wallet_id is not None:
            self.wallet_id = wallet_id
        self.workload_id = workload_id

    @property
    def additional_data(self):
        """Gets the additional_data of this OrmWalletHistory.  # noqa: E501


        :return: The additional_data of this OrmWalletHistory.  # noqa: E501
        :rtype: dict[str, object]
        """
        return self._additional_data

    @additional_data.setter
    def additional_data(self, additional_data):
        """Sets the additional_data of this OrmWalletHistory.


        :param additional_data: The additional_data of this OrmWalletHistory.  # noqa: E501
        :type additional_data: dict[str, object]
        """

        self._additional_data = additional_data

    @property
    def artifact_id(self):
        """Gets the artifact_id of this OrmWalletHistory.  # noqa: E501


        :return: The artifact_id of this OrmWalletHistory.  # noqa: E501
        :rtype: int
        """
        return self._artifact_id

    @artifact_id.setter
    def artifact_id(self, artifact_id):
        """Sets the artifact_id of this OrmWalletHistory.


        :param artifact_id: The artifact_id of this OrmWalletHistory.  # noqa: E501
        :type artifact_id: int
        """

        self._artifact_id = artifact_id

    @property
    def balance_from(self):
        """Gets the balance_from of this OrmWalletHistory.  # noqa: E501


        :return: The balance_from of this OrmWalletHistory.  # noqa: E501
        :rtype: int
        """
        return self._balance_from

    @balance_from.setter
    def balance_from(self, balance_from):
        """Sets the balance_from of this OrmWalletHistory.


        :param balance_from: The balance_from of this OrmWalletHistory.  # noqa: E501
        :type balance_from: int
        """

        self._balance_from = balance_from

    @property
    def balance_to(self):
        """Gets the balance_to of this OrmWalletHistory.  # noqa: E501


        :return: The balance_to of this OrmWalletHistory.  # noqa: E501
        :rtype: int
        """
        return self._balance_to

    @balance_to.setter
    def balance_to(self, balance_to):
        """Sets the balance_to of this OrmWalletHistory.


        :param balance_to: The balance_to of this OrmWalletHistory.  # noqa: E501
        :type balance_to: int
        """

        self._balance_to = balance_to

    @property
    def cluster_id(self):
        """Gets the cluster_id of this OrmWalletHistory.  # noqa: E501


        :return: The cluster_id of this OrmWalletHistory.  # noqa: E501
        :rtype: int
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id):
        """Sets the cluster_id of this OrmWalletHistory.


        :param cluster_id: The cluster_id of this OrmWalletHistory.  # noqa: E501
        :type cluster_id: int
        """

        self._cluster_id = cluster_id

    @property
    def coupon_redemption_id(self):
        """Gets the coupon_redemption_id of this OrmWalletHistory.  # noqa: E501


        :return: The coupon_redemption_id of this OrmWalletHistory.  # noqa: E501
        :rtype: int
        """
        return self._coupon_redemption_id

    @coupon_redemption_id.setter
    def coupon_redemption_id(self, coupon_redemption_id):
        """Sets the coupon_redemption_id of this OrmWalletHistory.


        :param coupon_redemption_id: The coupon_redemption_id of this OrmWalletHistory.  # noqa: E501
        :type coupon_redemption_id: int
        """

        self._coupon_redemption_id = coupon_redemption_id

    @property
    def created_dt(self):
        """Gets the created_dt of this OrmWalletHistory.  # noqa: E501


        :return: The created_dt of this OrmWalletHistory.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this OrmWalletHistory.


        :param created_dt: The created_dt of this OrmWalletHistory.  # noqa: E501
        :type created_dt: datetime
        """

        self._created_dt = created_dt

    @property
    def edges(self):
        """Gets the edges of this OrmWalletHistory.  # noqa: E501


        :return: The edges of this OrmWalletHistory.  # noqa: E501
        :rtype: OrmWalletHistoryEdges
        """
        return self._edges

    @edges.setter
    def edges(self, edges):
        """Sets the edges of this OrmWalletHistory.


        :param edges: The edges of this OrmWalletHistory.  # noqa: E501
        :type edges: OrmWalletHistoryEdges
        """

        self._edges = edges

    @property
    def id(self):
        """Gets the id of this OrmWalletHistory.  # noqa: E501


        :return: The id of this OrmWalletHistory.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this OrmWalletHistory.


        :param id: The id of this OrmWalletHistory.  # noqa: E501
        :type id: int
        """

        self._id = id

    @property
    def organization_id(self):
        """Gets the organization_id of this OrmWalletHistory.  # noqa: E501


        :return: The organization_id of this OrmWalletHistory.  # noqa: E501
        :rtype: int
        """
        return self._organization_id

    @organization_id.setter
    def organization_id(self, organization_id):
        """Sets the organization_id of this OrmWalletHistory.


        :param organization_id: The organization_id of this OrmWalletHistory.  # noqa: E501
        :type organization_id: int
        """

        self._organization_id = organization_id

    @property
    def reason(self):
        """Gets the reason of this OrmWalletHistory.  # noqa: E501


        :return: The reason of this OrmWalletHistory.  # noqa: E501
        :rtype: str
        """
        return self._reason

    @reason.setter
    def reason(self, reason):
        """Sets the reason of this OrmWalletHistory.


        :param reason: The reason of this OrmWalletHistory.  # noqa: E501
        :type reason: str
        """

        self._reason = reason

    @property
    def type(self):
        """Gets the type of this OrmWalletHistory.  # noqa: E501


        :return: The type of this OrmWalletHistory.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this OrmWalletHistory.


        :param type: The type of this OrmWalletHistory.  # noqa: E501
        :type type: str
        """

        self._type = type

    @property
    def updated_dt(self):
        """Gets the updated_dt of this OrmWalletHistory.  # noqa: E501


        :return: The updated_dt of this OrmWalletHistory.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this OrmWalletHistory.


        :param updated_dt: The updated_dt of this OrmWalletHistory.  # noqa: E501
        :type updated_dt: datetime
        """

        self._updated_dt = updated_dt

    @property
    def wallet_id(self):
        """Gets the wallet_id of this OrmWalletHistory.  # noqa: E501


        :return: The wallet_id of this OrmWalletHistory.  # noqa: E501
        :rtype: int
        """
        return self._wallet_id

    @wallet_id.setter
    def wallet_id(self, wallet_id):
        """Sets the wallet_id of this OrmWalletHistory.


        :param wallet_id: The wallet_id of this OrmWalletHistory.  # noqa: E501
        :type wallet_id: int
        """

        self._wallet_id = wallet_id

    @property
    def workload_id(self):
        """Gets the workload_id of this OrmWalletHistory.  # noqa: E501


        :return: The workload_id of this OrmWalletHistory.  # noqa: E501
        :rtype: int
        """
        return self._workload_id

    @workload_id.setter
    def workload_id(self, workload_id):
        """Sets the workload_id of this OrmWalletHistory.


        :param workload_id: The workload_id of this OrmWalletHistory.  # noqa: E501
        :type workload_id: int
        """

        self._workload_id = workload_id

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmWalletHistory):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmWalletHistory):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class HubModelTaskCreateAPIInput(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'category_id': 'int',
        'description': 'str',
        'key': 'str',
        'order_number': 'int',
        'overview': 'str',
        'scope': 'str',
        'status': 'str',
        'tags': 'list[ProtoTag]',
        'thumbnail_url': 'str',
        'title': 'str',
        'type': 'str',
        'yaml': 'str'
    }

    attribute_map = {
        'category_id': 'category_id',
        'description': 'description',
        'key': 'key',
        'order_number': 'order_number',
        'overview': 'overview',
        'scope': 'scope',
        'status': 'status',
        'tags': 'tags',
        'thumbnail_url': 'thumbnail_url',
        'title': 'title',
        'type': 'type',
        'yaml': 'yaml'
    }

    def __init__(self, category_id=None, description=None, key=None, order_number=None, overview=None, scope='private', status=None, tags=None, thumbnail_url=None, title=None, type=None, yaml=None, local_vars_configuration=None):  # noqa: E501
        """HubModelTaskCreateAPIInput - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._category_id = None
        self._description = None
        self._key = None
        self._order_number = None
        self._overview = None
        self._scope = None
        self._status = None
        self._tags = None
        self._thumbnail_url = None
        self._title = None
        self._type = None
        self._yaml = None
        self.discriminator = None

        self.category_id = category_id
        self.description = description
        self.key = key
        self.order_number = order_number
        self.overview = overview
        if scope is not None:
            self.scope = scope
        self.status = status
        if tags is not None:
            self.tags = tags
        self.thumbnail_url = thumbnail_url
        self.title = title
        self.type = type
        if yaml is not None:
            self.yaml = yaml

    @property
    def category_id(self):
        """Gets the category_id of this HubModelTaskCreateAPIInput.  # noqa: E501


        :return: The category_id of this HubModelTaskCreateAPIInput.  # noqa: E501
        :rtype: int
        """
        return self._category_id

    @category_id.setter
    def category_id(self, category_id):
        """Sets the category_id of this HubModelTaskCreateAPIInput.


        :param category_id: The category_id of this HubModelTaskCreateAPIInput.  # noqa: E501
        :type category_id: int
        """

        self._category_id = category_id

    @property
    def description(self):
        """Gets the description of this HubModelTaskCreateAPIInput.  # noqa: E501


        :return: The description of this HubModelTaskCreateAPIInput.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this HubModelTaskCreateAPIInput.


        :param description: The description of this HubModelTaskCreateAPIInput.  # noqa: E501
        :type description: str
        """

        self._description = description

    @property
    def key(self):
        """Gets the key of this HubModelTaskCreateAPIInput.  # noqa: E501


        :return: The key of this HubModelTaskCreateAPIInput.  # noqa: E501
        :rtype: str
        """
        return self._key

    @key.setter
    def key(self, key):
        """Sets the key of this HubModelTaskCreateAPIInput.


        :param key: The key of this HubModelTaskCreateAPIInput.  # noqa: E501
        :type key: str
        """

        self._key = key

    @property
    def order_number(self):
        """Gets the order_number of this HubModelTaskCreateAPIInput.  # noqa: E501


        :return: The order_number of this HubModelTaskCreateAPIInput.  # noqa: E501
        :rtype: int
        """
        return self._order_number

    @order_number.setter
    def order_number(self, order_number):
        """Sets the order_number of this HubModelTaskCreateAPIInput.


        :param order_number: The order_number of this HubModelTaskCreateAPIInput.  # noqa: E501
        :type order_number: int
        """

        self._order_number = order_number

    @property
    def overview(self):
        """Gets the overview of this HubModelTaskCreateAPIInput.  # noqa: E501


        :return: The overview of this HubModelTaskCreateAPIInput.  # noqa: E501
        :rtype: str
        """
        return self._overview

    @overview.setter
    def overview(self, overview):
        """Sets the overview of this HubModelTaskCreateAPIInput.


        :param overview: The overview of this HubModelTaskCreateAPIInput.  # noqa: E501
        :type overview: str
        """

        self._overview = overview

    @property
    def scope(self):
        """Gets the scope of this HubModelTaskCreateAPIInput.  # noqa: E501


        :return: The scope of this HubModelTaskCreateAPIInput.  # noqa: E501
        :rtype: str
        """
        return self._scope

    @scope.setter
    def scope(self, scope):
        """Sets the scope of this HubModelTaskCreateAPIInput.


        :param scope: The scope of this HubModelTaskCreateAPIInput.  # noqa: E501
        :type scope: str
        """
        allowed_values = ["private", "organization", "public"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and scope not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `scope` ({0}), must be one of {1}"  # noqa: E501
                .format(scope, allowed_values)
            )

        self._scope = scope

    @property
    def status(self):
        """Gets the status of this HubModelTaskCreateAPIInput.  # noqa: E501


        :return: The status of this HubModelTaskCreateAPIInput.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this HubModelTaskCreateAPIInput.


        :param status: The status of this HubModelTaskCreateAPIInput.  # noqa: E501
        :type status: str
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501
        allowed_values = ["draft", "published"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and status not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `status` ({0}), must be one of {1}"  # noqa: E501
                .format(status, allowed_values)
            )

        self._status = status

    @property
    def tags(self):
        """Gets the tags of this HubModelTaskCreateAPIInput.  # noqa: E501


        :return: The tags of this HubModelTaskCreateAPIInput.  # noqa: E501
        :rtype: list[ProtoTag]
        """
        return self._tags

    @tags.setter
    def tags(self, tags):
        """Sets the tags of this HubModelTaskCreateAPIInput.


        :param tags: The tags of this HubModelTaskCreateAPIInput.  # noqa: E501
        :type tags: list[ProtoTag]
        """

        self._tags = tags

    @property
    def thumbnail_url(self):
        """Gets the thumbnail_url of this HubModelTaskCreateAPIInput.  # noqa: E501


        :return: The thumbnail_url of this HubModelTaskCreateAPIInput.  # noqa: E501
        :rtype: str
        """
        return self._thumbnail_url

    @thumbnail_url.setter
    def thumbnail_url(self, thumbnail_url):
        """Sets the thumbnail_url of this HubModelTaskCreateAPIInput.


        :param thumbnail_url: The thumbnail_url of this HubModelTaskCreateAPIInput.  # noqa: E501
        :type thumbnail_url: str
        """

        self._thumbnail_url = thumbnail_url

    @property
    def title(self):
        """Gets the title of this HubModelTaskCreateAPIInput.  # noqa: E501


        :return: The title of this HubModelTaskCreateAPIInput.  # noqa: E501
        :rtype: str
        """
        return self._title

    @title.setter
    def title(self, title):
        """Sets the title of this HubModelTaskCreateAPIInput.


        :param title: The title of this HubModelTaskCreateAPIInput.  # noqa: E501
        :type title: str
        """
        if self.local_vars_configuration.client_side_validation and title is None:  # noqa: E501
            raise ValueError("Invalid value for `title`, must not be `None`")  # noqa: E501

        self._title = title

    @property
    def type(self):
        """Gets the type of this HubModelTaskCreateAPIInput.  # noqa: E501


        :return: The type of this HubModelTaskCreateAPIInput.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this HubModelTaskCreateAPIInput.


        :param type: The type of this HubModelTaskCreateAPIInput.  # noqa: E501
        :type type: str
        """
        if self.local_vars_configuration.client_side_validation and type is None:  # noqa: E501
            raise ValueError("Invalid value for `type`, must not be `None`")  # noqa: E501
        allowed_values = ["run", "service", "serverless", "pipeline"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and type not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `type` ({0}), must be one of {1}"  # noqa: E501
                .format(type, allowed_values)
            )

        self._type = type

    @property
    def yaml(self):
        """Gets the yaml of this HubModelTaskCreateAPIInput.  # noqa: E501


        :return: The yaml of this HubModelTaskCreateAPIInput.  # noqa: E501
        :rtype: str
        """
        return self._yaml

    @yaml.setter
    def yaml(self, yaml):
        """Sets the yaml of this HubModelTaskCreateAPIInput.


        :param yaml: The yaml of this HubModelTaskCreateAPIInput.  # noqa: E501
        :type yaml: str
        """

        self._yaml = yaml

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, HubModelTaskCreateAPIInput):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, HubModelTaskCreateAPIInput):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class ResponseActiveModelServiceRolloutSummary(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'current_step': 'int',
        'current_step_description': 'str',
        'expected_final_revisions': 'list[OrmExpectedRevisionStatus]',
        'is_rollback': 'bool',
        'message': 'str',
        'total_steps': 'int'
    }

    attribute_map = {
        'current_step': 'current_step',
        'current_step_description': 'current_step_description',
        'expected_final_revisions': 'expected_final_revisions',
        'is_rollback': 'is_rollback',
        'message': 'message',
        'total_steps': 'total_steps'
    }

    def __init__(self, current_step=None, current_step_description=None, expected_final_revisions=None, is_rollback=None, message=None, total_steps=None, local_vars_configuration=None):  # noqa: E501
        """ResponseActiveModelServiceRolloutSummary - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._current_step = None
        self._current_step_description = None
        self._expected_final_revisions = None
        self._is_rollback = None
        self._message = None
        self._total_steps = None
        self.discriminator = None

        self.current_step = current_step
        self.current_step_description = current_step_description
        self.expected_final_revisions = expected_final_revisions
        self.is_rollback = is_rollback
        self.message = message
        self.total_steps = total_steps

    @property
    def current_step(self):
        """Gets the current_step of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501


        :return: The current_step of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501
        :rtype: int
        """
        return self._current_step

    @current_step.setter
    def current_step(self, current_step):
        """Sets the current_step of this ResponseActiveModelServiceRolloutSummary.


        :param current_step: The current_step of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501
        :type current_step: int
        """
        if self.local_vars_configuration.client_side_validation and current_step is None:  # noqa: E501
            raise ValueError("Invalid value for `current_step`, must not be `None`")  # noqa: E501

        self._current_step = current_step

    @property
    def current_step_description(self):
        """Gets the current_step_description of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501


        :return: The current_step_description of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501
        :rtype: str
        """
        return self._current_step_description

    @current_step_description.setter
    def current_step_description(self, current_step_description):
        """Sets the current_step_description of this ResponseActiveModelServiceRolloutSummary.


        :param current_step_description: The current_step_description of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501
        :type current_step_description: str
        """
        if self.local_vars_configuration.client_side_validation and current_step_description is None:  # noqa: E501
            raise ValueError("Invalid value for `current_step_description`, must not be `None`")  # noqa: E501

        self._current_step_description = current_step_description

    @property
    def expected_final_revisions(self):
        """Gets the expected_final_revisions of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501


        :return: The expected_final_revisions of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501
        :rtype: list[OrmExpectedRevisionStatus]
        """
        return self._expected_final_revisions

    @expected_final_revisions.setter
    def expected_final_revisions(self, expected_final_revisions):
        """Sets the expected_final_revisions of this ResponseActiveModelServiceRolloutSummary.


        :param expected_final_revisions: The expected_final_revisions of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501
        :type expected_final_revisions: list[OrmExpectedRevisionStatus]
        """
        if self.local_vars_configuration.client_side_validation and expected_final_revisions is None:  # noqa: E501
            raise ValueError("Invalid value for `expected_final_revisions`, must not be `None`")  # noqa: E501

        self._expected_final_revisions = expected_final_revisions

    @property
    def is_rollback(self):
        """Gets the is_rollback of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501


        :return: The is_rollback of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501
        :rtype: bool
        """
        return self._is_rollback

    @is_rollback.setter
    def is_rollback(self, is_rollback):
        """Sets the is_rollback of this ResponseActiveModelServiceRolloutSummary.


        :param is_rollback: The is_rollback of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501
        :type is_rollback: bool
        """
        if self.local_vars_configuration.client_side_validation and is_rollback is None:  # noqa: E501
            raise ValueError("Invalid value for `is_rollback`, must not be `None`")  # noqa: E501

        self._is_rollback = is_rollback

    @property
    def message(self):
        """Gets the message of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501


        :return: The message of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """Sets the message of this ResponseActiveModelServiceRolloutSummary.


        :param message: The message of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501
        :type message: str
        """
        if self.local_vars_configuration.client_side_validation and message is None:  # noqa: E501
            raise ValueError("Invalid value for `message`, must not be `None`")  # noqa: E501

        self._message = message

    @property
    def total_steps(self):
        """Gets the total_steps of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501


        :return: The total_steps of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501
        :rtype: int
        """
        return self._total_steps

    @total_steps.setter
    def total_steps(self, total_steps):
        """Sets the total_steps of this ResponseActiveModelServiceRolloutSummary.


        :param total_steps: The total_steps of this ResponseActiveModelServiceRolloutSummary.  # noqa: E501
        :type total_steps: int
        """
        if self.local_vars_configuration.client_side_validation and total_steps is None:  # noqa: E501
            raise ValueError("Invalid value for `total_steps`, must not be `None`")  # noqa: E501

        self._total_steps = total_steps

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseActiveModelServiceRolloutSummary):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseActiveModelServiceRolloutSummary):
            return True

        return self.to_dict() != other.to_dict()



from vessl.util.constant import WEB_HOST


class Endpoint:
    cluster = WEB_HOST + "/{}/clusters/{}"  # Orgainzation name, cluster ID
    dataset = WEB_HOST + "/{}/datasets/{}"  # Organization name, dataset name
    experiment = (
        WEB_HOST + "/{}/{}/experiments/{}"
    )  # Organization name, project name, experiment number
    experiment_logs = (
        WEB_HOST + "/{}/{}/experiments/{}/logs"
    )  # Organization name, project name, experiment number
    run = WEB_HOST + "/{}/runs/{}/{}"
    model_repository = WEB_HOST + "/{}/models/{}"  # Organization name, repository name
    model = WEB_HOST + "/{}/models/{}/{}"  # Organization name, repository name, number
    organization = WEB_HOST + "/{}"  # Organization name
    project = WEB_HOST + "/{}/{}"  # Organization name, project name
    sweep = WEB_HOST + "/{}/{}/sweeps/{}"  # Organization name, project name, sweep name
    sweep_logs = WEB_HOST + "/{}/{}/sweeps/{}/logs"  # Organization name, project name, sweep name
    workspace = WEB_HOST + "/{}/workspaces/{}"  # Organization name, workspace id
    service = WEB_HOST + "/{}/services/{}"  # Organization name, service name



import base64
import fnmatch
import json
import os
from typing import List, Tuple

import boto3
import botocore.client
from boto3.s3.transfer import S3Transfer
from botocore.credentials import Credentials, RefreshableCredentials
from botocore.session import Session as BotocoreSession
from google.cloud import storage
from google.oauth2.service_account import Credentials as GoogleServiceAccountCredentials

from vessl import vessl_api
from vessl.util.fmt import sizeof_fmt
from vessl.util.common import remove_prefix, safe_cast
from vessl.util.constant import VESSL_IGNORE_FILE_PATH
from vessl.util.downloader import Downloader
from vessl.util.exception import InvalidParamsError, InvalidVolumeFileError
from vessl.util.google_storage_uploader import GoogleStorageUploader


class FileTransmission:
    def __init__(self, source_path, source_abs_path, dest_abs_path, size):
        self.source_path = source_path
        self.source_abs_path = source_abs_path
        self.dest_abs_path = dest_abs_path
        self.size = size


class VolumeFileTransfer:
    def __init__(self, volume_id: int):
        self.volume = vessl_api.volume_read_api(volume_id=volume_id)
        self.provider = None
        self.region_name = None
        self.bucket_name = None
        self.prefix = None
        self.endpoint = None
        self.s3_client = None
        self.gs_client = None
        self._update_federation_credentials()

    def download(self, source_path, dest_path):
        if self.provider == "gs":
            result = vessl_api.volume_file_list_api(
                volume_id=self.volume.id,
                recursive=True,
                path=source_path,
                need_download_url=True,
            ).results
            files = sorted(result, key=lambda x: x.path)
            Downloader.download(source_path, dest_path, *files, quiet=True)
            return

        (
            file_transmissions,
            total_size,
        ) = self._get_download_file_transmissions_and_total_size(source_path, dest_path)
        total_file_count = len(file_transmissions)
        if total_file_count == 0:
            print("No files to download.")
            return

        formatted_total_size = sizeof_fmt(total_size)
        print(f"Downloading {total_file_count} file(s) ({formatted_total_size})...")

        succeeded_count = 0
        succeeded_size = 0
        for file_transmission in file_transmissions:
            dirname = os.path.dirname(file_transmission.dest_abs_path)
            if dirname:
                os.makedirs(dirname, exist_ok=True)
            try:
                self._get_s3_client().download_file(
                    self.bucket_name,
                    file_transmission.source_abs_path,
                    file_transmission.dest_abs_path,
                    Callback=self._create_callback(succeeded_size, total_size),
                )
                succeeded_count += 1
                succeeded_size += file_transmission.size
            except BaseException as e:
                print(f"Failed to download {file_transmission.source_path}.")
                raise e

        print(f"Total {succeeded_count} file(s) downloaded.")

    def upload(self, source_path, dest_path):
        if self.volume.is_read_only:
            print("Cannot upload to read-only volume.")
            return

        ignores = []
        if os.path.exists(VESSL_IGNORE_FILE_PATH):
            with open(VESSL_IGNORE_FILE_PATH, "r") as f:
                ignores = [line.strip() for line in f.readlines()]

        if self.provider == "gs":
            client = self._get_gs_client()
            GoogleStorageUploader.upload(
                client, source_path, self.bucket_name, self.prefix, ignores
            )
            return

        s3_client = self._get_s3_client()
        (
            file_transmissions,
            total_size,
        ) = self._get_upload_file_transmissions_and_total_size(source_path, dest_path, ignores)
        total_file_count = len(file_transmissions)
        if total_file_count == 0:
            print("No files to upload.")
            return

        formatted_total_size = sizeof_fmt(total_size)
        print(f"Uploading {total_file_count} file(s) ({formatted_total_size})...")

        succeeded_count = 0
        succeeded_size = 0
        succeeded_files = []
        for file_transmission in file_transmissions:
            try:
                s3_client.upload_file(
                    file_transmission.source_abs_path,
                    self.bucket_name,
                    file_transmission.dest_abs_path,
                    Callback=self._create_callback(succeeded_size, total_size),
                )
                succeeded_count += 1
                succeeded_size += file_transmission.size
                succeeded_files.append({"path": os.path.basename(file_transmission.dest_abs_path)})
            except BaseException as e:
                print(f"Failed to upload {file_transmission.source_path}.")
                raise e

        print(f"Total {succeeded_count} file(s) uploaded.")
        return succeeded_files

    def copy(self, source_path, dest_path):
        # TODO
        return

    def remove(self, path):
        # TODO
        return

    def _create_callback(self, current_size, total_size):
        if total_size < 100 * 1024 * 1024:
            return None

        total_transmitted = current_size
        last_percent = int(total_transmitted / total_size * 100)
        formatted_size = sizeof_fmt(total_size)
        interval = max(int(31 * 1024 * 1024 / total_size * 100), 3)

        def callback(transmitted_bytes):
            nonlocal total_transmitted, last_percent
            total_transmitted += transmitted_bytes
            percent = int(total_transmitted / total_size * 100)
            if percent - last_percent >= interval or percent == 100:
                print(f"{sizeof_fmt(total_transmitted)}/{formatted_size} ({percent}%) completed.")
                last_percent = percent

        return callback

    def _update_federation_credentials(self):
        federation_credentials = vessl_api.volume_federate_api(volume_id=self.volume.id)
        self.region_name = federation_credentials.region
        self.bucket_name = federation_credentials.bucket
        self.prefix = federation_credentials.prefix
        self.provider = federation_credentials.token.provider
        self.endpoint = federation_credentials.endpoint

    def _get_gs_client(self):
        if self.provider != "gs":
            raise InvalidVolumeFileError("This volume is not a Google Cloud Storage volume.")
        if self.gs_client:
            return self.gs_client

        self.gs_client = storage.Client(credentials=self.__get_gs_downscoped_credentials())
        return self.gs_client

    def __get_s3_api_endpoint_url(self):
        return os.environ.get("VESSL_VOLUME_S3_API_ENDPOINT_URL", self.endpoint)

    def __get_s3_api_verify_ssl(self):
        insecure_skip_tls_verify = safe_cast(os.environ.get('VESSL_INSECURE_SKIP_TLS_VERIFY'), bool, False)
        verify_tls = not insecure_skip_tls_verify
        if verify_tls:
            return True

        if os.environ.get("VESSL_VOLUME_S3_API_VERIFY_SSL") == "false":
            return False
        return None

    def _get_s3_client(self) -> S3Transfer:
        if self.s3_client:
            return self.s3_client

        no_refresh = os.environ.get("VESSL_VOLUME_CREDENTIAL_NO_REFRESH") == "true"
        if no_refresh:
            return self._get_s3_client_from_static_credentials()

        botocore_session = BotocoreSession()
        botocore_session.set_config_variable("region", self.region_name)

        creds = self.__get_s3_session_credentials()
        if creds is not None:
            refreshable_credentials = RefreshableCredentials.create_from_metadata(
                metadata=creds,
                refresh_using=self.__get_s3_session_credentials,
                method="sts-assume-role",
            )
            botocore_session._credentials = refreshable_credentials

        session = boto3.session.Session(botocore_session=botocore_session)
        config = botocore.client.Config(connect_timeout=120, read_timeout=120)
        self.s3_client = session.client(
            "s3",
            config=config,
            endpoint_url=self.__get_s3_api_endpoint_url(),
            verify=self.__get_s3_api_verify_ssl(),
        )
        return self.s3_client

    def _get_s3_client_from_static_credentials(self) -> S3Transfer:
        if self.s3_client:
            return self.s3_client

        metadata = self.__get_s3_session_credentials()
        credential = Credentials(
            access_key=metadata["access_key"], secret_key=metadata["secret_key"]
        )
        botocore_session = BotocoreSession()
        botocore_session._credentials = credential
        botocore_session.set_config_variable("region", self.region_name)
        session = boto3.session.Session(botocore_session=botocore_session)
        config = botocore.client.Config(connect_timeout=120, read_timeout=120)
        self.s3_client = session.client(
            "s3",
            config=config,
            endpoint_url=self.__get_s3_api_endpoint_url(),
            verify=self.__get_s3_api_verify_ssl(),
        )
        return self.s3_client

    def __get_all_s3_objects(self, **base_kwargs):
        continuation_token = None
        while True:
            list_kwargs = dict(MaxKeys=1000, **base_kwargs)
            if continuation_token:
                list_kwargs["ContinuationToken"] = continuation_token

            response = self._get_s3_client().list_objects_v2(**list_kwargs)
            yield from filter(
                lambda x: not x["Key"].endswith("/") or x["Size"] > 0, response.get("Contents", [])
            )

            if not response.get("IsTruncated"):
                break

            continuation_token = response.get("NextContinuationToken")

    def _get_download_file_transmissions_and_total_size(
        self, source_path: str, dest_path: str
    ) -> Tuple[List[FileTransmission], int]:
        source_path = source_path.strip("/")
        contents_only = source_path == "." or source_path.endswith("/.")
        if contents_only:
            source_path = source_path[:-2] if source_path.endswith("/.") else ""
        is_root = source_path == ""
        source_abs_path = (
            os.path.join(self.prefix, source_path).replace(os.sep, "/").rstrip("/")
            if self.prefix
            else source_path
        )
        dest_abs_path = os.path.abspath(dest_path)

        objects = list(self.__get_all_s3_objects(Bucket=self.bucket_name, Prefix=source_abs_path))
        if len(objects) == 0:
            return [], 0

        if len(objects) > 1 and os.path.exists(dest_abs_path) and not os.path.isdir(dest_abs_path):
            raise InvalidParamsError(f"{dest_abs_path} is not a directory.")

        if len(objects) == 1 and source_path == remove_prefix(objects[0]["Key"], self.prefix + "/"):
            file_transmission = FileTransmission(
                source_path=source_path,
                source_abs_path=objects[0]["Key"],
                dest_abs_path=dest_abs_path,
                size=objects[0]["Size"],
            )
            if os.path.isdir(dest_abs_path):
                file_transmission.dest_abs_path = os.path.join(
                    dest_abs_path, source_path.split("/")[-1]
                )

            return [file_transmission], file_transmission.size

        prefix = source_abs_path if contents_only else source_abs_path.rsplit("/", 1)[0]
        total_size = 0
        file_transmissions = []
        for obj in objects:
            total_size += obj["Size"]
            if is_root:
                source_rel_path = remove_prefix(obj["Key"], self.prefix + "/")
            else:
                source_rel_path = remove_prefix(obj["Key"], prefix + "/")

            file_transmissions.append(
                FileTransmission(
                    source_path=source_rel_path,
                    source_abs_path=obj["Key"],
                    dest_abs_path=os.path.join(dest_abs_path, source_rel_path),
                    size=obj["Size"],
                )
            )

        return file_transmissions, total_size

    def _get_upload_file_transmissions_and_total_size(
        self, source_path: str, dest_path: str, ignores: List[str] = []
    ) -> Tuple[List[FileTransmission], int]:
        contents_only = source_path.endswith("/.")
        source_path = os.path.relpath(source_path)
        if not os.path.exists(source_path):
            raise InvalidParamsError(f"{source_path} does not exist.")

        dest_path = dest_path.strip("/")
        dest_abs_path = (
            os.path.join(self.prefix, dest_path).replace(os.sep, "/").rstrip("/")
            if self.prefix
            else dest_path
        )

        s3_client = self._get_s3_client()
        exist = True
        is_root = dest_path == ""
        is_dir = is_root
        try:
            s3_client.head_object(Bucket=self.bucket_name, Key=dest_abs_path)
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "404":
                response = s3_client.list_objects_v2(Bucket=self.bucket_name, Prefix=dest_abs_path)
                if response["KeyCount"] == 0:
                    exist = False
                else:
                    is_dir = True
            else:
                raise

        source_file_name = os.path.basename(source_path)
        source_abs_path = os.path.abspath(source_path)
        if os.path.isfile(source_path):
            if len(ignores) > 0 and any(
                fnmatch.fnmatch(source_file_name, ignore) for ignore in ignores
            ):
                return [], 0
            if not exist or not is_dir:
                file_transmission = FileTransmission(
                    source_path=source_path,
                    source_abs_path=source_abs_path,
                    dest_abs_path=dest_abs_path,
                    size=os.path.getsize(source_abs_path),
                )
                return [file_transmission], file_transmission.size
            else:
                file_transmission = FileTransmission(
                    source_path=source_path,
                    source_abs_path=source_abs_path,
                    dest_abs_path=os.path.join(dest_abs_path, source_file_name).replace(
                        os.sep, "/"
                    ),
                    size=os.path.getsize(source_abs_path),
                )
                return [file_transmission], file_transmission.size

        elif not exist or is_dir:
            total_size = 0
            file_transmissions = []
            for root_path, dirs, file_names in os.walk(source_path, topdown=True):
                # Remove ignored directories
                dirs[:] = [
                    d for d in dirs if not any(fnmatch.fnmatch(d, ignore) for ignore in ignores)
                ]
                for file_name in file_names:
                    if any(fnmatch.fnmatch(file_name, ignore) for ignore in ignores):
                        continue
                    source_file_path = os.path.join(root_path, file_name)
                    source_file_abs_path = os.path.abspath(source_file_path)
                    prefix = source_abs_path if contents_only else source_abs_path.rsplit("/", 1)[0]
                    source_rel_path = remove_prefix(source_file_abs_path, prefix + "/")

                    file_size = os.path.getsize(source_abs_path)
                    file_transmissions.append(
                        FileTransmission(
                            source_file_path,
                            source_file_abs_path,
                            os.path.join(dest_abs_path, source_rel_path).replace(os.sep, "/"),
                            file_size,
                        )
                    )
                    total_size += file_size

            return file_transmissions, total_size
        else:
            raise InvalidParamsError(f"{dest_path} is not a directory.")

    def __get_s3_session_credentials(self):
        federation_credentials = vessl_api.volume_federate_api(volume_id=self.volume.id)
        if federation_credentials.token.s3 is None:
            return None

        creds = federation_credentials.token.s3
        return {
            "access_key": creds.access_key_id,
            "secret_key": creds.secret_access_key,
            "token": creds.session_token,
            "expiry_time": creds.expiration.isoformat(),
        }

    def __get_gs_downscoped_credentials(self):
        federation_credentials = vessl_api.volume_federate_api(volume_id=self.volume.id)
        json_key_bytes = base64.b64decode(federation_credentials.token.gs.base64_credentials)
        json_key = json.loads(json_key_bytes)
        return GoogleServiceAccountCredentials.from_service_account_info(json_key)



from typing import Any

import click

from vessl import VESSL_ENV


def print_warning(text: str, stderr: bool = True) -> Any:
    click.echo(style_warning(text), err=stderr)


def style_warning(text: str) -> Any:
    return click.style(text, fg="magenta")


def print_debug(text: str, stderr: bool = True) -> Any:
    if VESSL_ENV == "dev":
        click.echo(f"DEBUG>> {style_info(text)}", err=stderr)


def print_info(text: str, stderr: bool = True) -> Any:
    click.echo(style_info(text), err=stderr)


def print_prompt_style(text: str, stderr: bool = True) -> Any:
    """
    Prints prompt style text not a prompt itself. Use for multi-line explanations on a prompt.
    """
    click.echo(style_prompt(text), err=stderr)


def print_success(text: str, stderr: bool = True) -> Any:
    click.echo(style_success(text), err=stderr)


def print_success_result(text: str, stderr: bool = True) -> Any:
    click.echo(style_success_result(text), err=stderr)


def print_error(text: str, stderr: bool = True) -> Any:
    click.echo(style_error(text), err=stderr)


def print_error_result(text: str, stderr: bool = True) -> Any:
    click.echo(style_error_result(text), err=stderr)


def style_info(text: str) -> Any:
    return click.style(text, fg="cyan")


def style_prompt(text: str) -> Any:
    return click.style(text, fg="yellow")


def style_error(text: str) -> Any:
    return click.style(text, fg="red")


def style_success(text: str) -> Any:
    return click.style(text, fg="green")


def style_success_result(text: str) -> Any:
    return click.style(text, bg="green")


def style_error_result(text: str) -> Any:
    return click.style(text, bg="red", fg="white")



from enum import Enum
from typing import Dict, Any

from vessl.openapi_client import ProtoLLMDocument


class DocumentParserType(Enum):
    OPENPARSE = "openparse"
    RAGFLOW = "ragflow"


class DocumentChunkingType(Enum):
    CHARACTER_SPLITTING = "character_splitting"
    FIXED_SIZE = "fixed_size"
    NAIVE_SPLITTING = "naive_splitting"
    NLTK = "nltk"
    SPACY = "spacy"


class Document:
    def __init__(self, document: ProtoLLMDocument):
        self.id: int = document.id
        self.filename: str = document.filename
        self.extension: str = document.extension
        self.size: str = document.size

        try:
            self.parser_type: DocumentParserType = DocumentParserType(document.parser_type)
        except ValueError:
            raise ValueError(f"Invalid parser_type: {document.parser_type}")

        try:
            self.chunking_type: DocumentChunkingType = DocumentChunkingType(document.chunking_type)
        except ValueError:
            raise ValueError(f"Invalid chunking_type: {document.chunking_type}")

        self.chunking_method_params: Dict[str, Any] = document.chunking_method_params



import os
from concurrent.futures import Future

from requests_futures.sessions import FuturesSession

from vessl.openapi_client import StorageFile
from vessl.util.common import remove_prefix
from vessl.util.file_transmission import (
    FileTransmissionHandler,
    FileTransmissionTracker,
)


class Downloader:
    @classmethod
    def download(
        cls,
        source_path: str,
        dest_path: str,
        *remote_files: StorageFile,
        quiet: bool = False,
    ):
        """Download volume files"""
        if not remote_files:
            return

        handler = FileTransmissionHandler(cls._download_future, action="download", quiet=quiet)
        if len(remote_files) == 1 and remote_files[0].path == source_path:
            handler.add_file(dest_path, remote_files[0].download_url.url, remote_files[0].size)
        else:
            for file in remote_files:
                file_path = remove_prefix(file.path, source_path).lstrip("/")
                dir_name = os.path.dirname(file_path)
                file_name = os.path.basename(file_path)
                local_full_path = os.path.join(dest_path, dir_name, file_name)
                handler.add_file(local_full_path, file.download_url.url, file.size)

        handler.run()

    @classmethod
    def _download_future(
        cls,
        session: FuturesSession,
        full_path: str,
        url: str,
        tracker: FileTransmissionTracker,
    ) -> Future:
        return session.get(
            url, stream=True, hooks={"response": cls._download_hook(full_path, tracker)}
        )

    @staticmethod
    def _download_hook(full_path, tracker=None):
        def fn(resp, **kwargs):
            if not resp.ok:
                return

            dirname = os.path.dirname(full_path)
            if dirname:
                os.makedirs(dirname, exist_ok=True)

            with open(full_path, "wb") as f:
                for chunk in resp.iter_content(chunk_size=8192):
                    f.write(chunk)
                    if tracker:
                        tracker.increase_done_size(len(chunk))
                        tracker.print_progress()
                if tracker:
                    tracker.increase_done_count()

        return fn



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class OrmLLMModelConnection(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'aws_region': 'str',
        'created_by_id': 'int',
        'created_dt': 'datetime',
        'description': 'str',
        'edges': 'OrmLLMModelConnectionEdges',
        'id': 'int',
        'name': 'str',
        'organization_id': 'int',
        'provider': 'str',
        'secret_id': 'int',
        'slug': 'str',
        'type': 'str',
        'updated_dt': 'datetime',
        'url': 'str'
    }

    attribute_map = {
        'aws_region': 'aws_region',
        'created_by_id': 'created_by_id',
        'created_dt': 'created_dt',
        'description': 'description',
        'edges': 'edges',
        'id': 'id',
        'name': 'name',
        'organization_id': 'organization_id',
        'provider': 'provider',
        'secret_id': 'secret_id',
        'slug': 'slug',
        'type': 'type',
        'updated_dt': 'updated_dt',
        'url': 'url'
    }

    def __init__(self, aws_region=None, created_by_id=None, created_dt=None, description=None, edges=None, id=None, name=None, organization_id=None, provider=None, secret_id=None, slug=None, type=None, updated_dt=None, url=None, local_vars_configuration=None):  # noqa: E501
        """OrmLLMModelConnection - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._aws_region = None
        self._created_by_id = None
        self._created_dt = None
        self._description = None
        self._edges = None
        self._id = None
        self._name = None
        self._organization_id = None
        self._provider = None
        self._secret_id = None
        self._slug = None
        self._type = None
        self._updated_dt = None
        self._url = None
        self.discriminator = None

        self.aws_region = aws_region
        if created_by_id is not None:
            self.created_by_id = created_by_id
        if created_dt is not None:
            self.created_dt = created_dt
        if description is not None:
            self.description = description
        if edges is not None:
            self.edges = edges
        if id is not None:
            self.id = id
        if name is not None:
            self.name = name
        if organization_id is not None:
            self.organization_id = organization_id
        if provider is not None:
            self.provider = provider
        self.secret_id = secret_id
        if slug is not None:
            self.slug = slug
        if type is not None:
            self.type = type
        if updated_dt is not None:
            self.updated_dt = updated_dt
        self.url = url

    @property
    def aws_region(self):
        """Gets the aws_region of this OrmLLMModelConnection.  # noqa: E501


        :return: The aws_region of this OrmLLMModelConnection.  # noqa: E501
        :rtype: str
        """
        return self._aws_region

    @aws_region.setter
    def aws_region(self, aws_region):
        """Sets the aws_region of this OrmLLMModelConnection.


        :param aws_region: The aws_region of this OrmLLMModelConnection.  # noqa: E501
        :type aws_region: str
        """

        self._aws_region = aws_region

    @property
    def created_by_id(self):
        """Gets the created_by_id of this OrmLLMModelConnection.  # noqa: E501


        :return: The created_by_id of this OrmLLMModelConnection.  # noqa: E501
        :rtype: int
        """
        return self._created_by_id

    @created_by_id.setter
    def created_by_id(self, created_by_id):
        """Sets the created_by_id of this OrmLLMModelConnection.


        :param created_by_id: The created_by_id of this OrmLLMModelConnection.  # noqa: E501
        :type created_by_id: int
        """

        self._created_by_id = created_by_id

    @property
    def created_dt(self):
        """Gets the created_dt of this OrmLLMModelConnection.  # noqa: E501


        :return: The created_dt of this OrmLLMModelConnection.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this OrmLLMModelConnection.


        :param created_dt: The created_dt of this OrmLLMModelConnection.  # noqa: E501
        :type created_dt: datetime
        """

        self._created_dt = created_dt

    @property
    def description(self):
        """Gets the description of this OrmLLMModelConnection.  # noqa: E501


        :return: The description of this OrmLLMModelConnection.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this OrmLLMModelConnection.


        :param description: The description of this OrmLLMModelConnection.  # noqa: E501
        :type description: str
        """

        self._description = description

    @property
    def edges(self):
        """Gets the edges of this OrmLLMModelConnection.  # noqa: E501


        :return: The edges of this OrmLLMModelConnection.  # noqa: E501
        :rtype: OrmLLMModelConnectionEdges
        """
        return self._edges

    @edges.setter
    def edges(self, edges):
        """Sets the edges of this OrmLLMModelConnection.


        :param edges: The edges of this OrmLLMModelConnection.  # noqa: E501
        :type edges: OrmLLMModelConnectionEdges
        """

        self._edges = edges

    @property
    def id(self):
        """Gets the id of this OrmLLMModelConnection.  # noqa: E501


        :return: The id of this OrmLLMModelConnection.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this OrmLLMModelConnection.


        :param id: The id of this OrmLLMModelConnection.  # noqa: E501
        :type id: int
        """

        self._id = id

    @property
    def name(self):
        """Gets the name of this OrmLLMModelConnection.  # noqa: E501


        :return: The name of this OrmLLMModelConnection.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this OrmLLMModelConnection.


        :param name: The name of this OrmLLMModelConnection.  # noqa: E501
        :type name: str
        """

        self._name = name

    @property
    def organization_id(self):
        """Gets the organization_id of this OrmLLMModelConnection.  # noqa: E501


        :return: The organization_id of this OrmLLMModelConnection.  # noqa: E501
        :rtype: int
        """
        return self._organization_id

    @organization_id.setter
    def organization_id(self, organization_id):
        """Sets the organization_id of this OrmLLMModelConnection.


        :param organization_id: The organization_id of this OrmLLMModelConnection.  # noqa: E501
        :type organization_id: int
        """

        self._organization_id = organization_id

    @property
    def provider(self):
        """Gets the provider of this OrmLLMModelConnection.  # noqa: E501


        :return: The provider of this OrmLLMModelConnection.  # noqa: E501
        :rtype: str
        """
        return self._provider

    @provider.setter
    def provider(self, provider):
        """Sets the provider of this OrmLLMModelConnection.


        :param provider: The provider of this OrmLLMModelConnection.  # noqa: E501
        :type provider: str
        """

        self._provider = provider

    @property
    def secret_id(self):
        """Gets the secret_id of this OrmLLMModelConnection.  # noqa: E501


        :return: The secret_id of this OrmLLMModelConnection.  # noqa: E501
        :rtype: int
        """
        return self._secret_id

    @secret_id.setter
    def secret_id(self, secret_id):
        """Sets the secret_id of this OrmLLMModelConnection.


        :param secret_id: The secret_id of this OrmLLMModelConnection.  # noqa: E501
        :type secret_id: int
        """

        self._secret_id = secret_id

    @property
    def slug(self):
        """Gets the slug of this OrmLLMModelConnection.  # noqa: E501


        :return: The slug of this OrmLLMModelConnection.  # noqa: E501
        :rtype: str
        """
        return self._slug

    @slug.setter
    def slug(self, slug):
        """Sets the slug of this OrmLLMModelConnection.


        :param slug: The slug of this OrmLLMModelConnection.  # noqa: E501
        :type slug: str
        """

        self._slug = slug

    @property
    def type(self):
        """Gets the type of this OrmLLMModelConnection.  # noqa: E501


        :return: The type of this OrmLLMModelConnection.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this OrmLLMModelConnection.


        :param type: The type of this OrmLLMModelConnection.  # noqa: E501
        :type type: str
        """

        self._type = type

    @property
    def updated_dt(self):
        """Gets the updated_dt of this OrmLLMModelConnection.  # noqa: E501


        :return: The updated_dt of this OrmLLMModelConnection.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this OrmLLMModelConnection.


        :param updated_dt: The updated_dt of this OrmLLMModelConnection.  # noqa: E501
        :type updated_dt: datetime
        """

        self._updated_dt = updated_dt

    @property
    def url(self):
        """Gets the url of this OrmLLMModelConnection.  # noqa: E501


        :return: The url of this OrmLLMModelConnection.  # noqa: E501
        :rtype: str
        """
        return self._url

    @url.setter
    def url(self, url):
        """Sets the url of this OrmLLMModelConnection.


        :param url: The url of this OrmLLMModelConnection.  # noqa: E501
        :type url: str
        """

        self._url = url

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmLLMModelConnection):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmLLMModelConnection):
            return True

        return self.to_dict() != other.to_dict()



import os
import time

import click

from vessl.util.exception import VesslApiException

from vessl.util.echo import print_warning, print_info

from vessl.cli._base import VesslGroup, VesslCommand
from vessl.cli.organization import organization_name_option
from vessl.pipeline import (
    get_paused_steps,
    resume_step,
    update_context_variables,
    list_pipeline_execution,
    pipeline_execution_abort,
    resume_step_api,
    format_step,
)
from vessl.util.prompt import prompt_confirm, prompt_choices
from vessl.util.ssh import ssh_command_from_endpoint, ssh_private_key_path_callback


@click.command(name="pipeline", cls=VesslGroup)
def cli():
    pass


@cli.vessl_command()
@organization_name_option
@click.option(
    "-p",
    "--key-path",
    type=click.STRING,
    help="SSH private key path.",
    callback=ssh_private_key_path_callback,
)
@click.option(
    "--pipeline-name", "--pipeline",
    type=click.STRING,
    help="Name of the pipeline.",
)
@click.option(
    "--number", "--execution",
    type=click.INT,
    help="Pipeline execution number.",
)
@click.option(
    "--step", "step_key",
    help="Step key.",
    type=click.STRING,
)
def ssh(key_path: str, pipeline_name: str, number: int, step_key: str):
    """This commands opens ssh debugging session to selected pipeline step."""
    steps = get_paused_steps(pipeline_name=pipeline_name, execution_number=number, step_key=step_key)
    choices = []
    for step in steps:
        step_name = format_step(step)
        command = ssh_command_from_endpoint(step.run_result.run_execution.endpoints.ssh.endpoint, key_path)
        choices.append((step_name, command))
    if len(choices) == 0:
        print_info("No matching idle pipeline execution found.")
        return
    elif len(choices) == 1:
        cmd = choices[0][1]
    else:
        cmd = prompt_choices("Choose step to ssh", choices)
    print_info(cmd)
    os.system(cmd)


@cli.vessl_command()
@organization_name_option
@click.option(
    "--pipeline-name", "--pipeline",
    type=click.STRING,
    help="Name of the pipeline.",
    required=True
)
@click.option(
    "--number", "--execution",
    help="Pipeline execution number.",
    type=click.INT,
)
@click.option(
    "--yes",
    is_flag=True,
    default=False,
)
def abort(pipeline_name: str, number: int, yes):
    """This commands aborts pipeline execution of pipeline_name."""
    try:
        execs = list_pipeline_execution(pipeline_name=pipeline_name)
    except VesslApiException as e:
        if e.code == 404:
            print_warning(f"Cannot find pipeline {pipeline_name}")
        else:
            print_warning(str(e))
        return

    if number is None:
        running = [e for e in execs if e.status == 'running']
        if len(running) == 0:
            print_warning(f"No execution running for {pipeline_name}")

        choices = []
        for e in running:
            choices.append((f"{pipeline_name} (#{e.number})", e))
        target = prompt_choices("Choose execution number to abort", choices)
        number = target.number

    if not yes:
        if not prompt_confirm(f"This will abort execution {pipeline_name} (#{number}). Are you sure?"):
            return

    pipeline_execution_abort(pipeline_name, number)
    print_info(f"{pipeline_name} (#{number}) has been aborted.")


@click.command(
    name="resume",
    cls=VesslCommand,
    login_required=True,
)
@organization_name_option
@click.option(
    "--pipeline-name", "--pipeline",
    type=click.STRING,
    help="Name of the pipeline.",
    required=True,
)
@click.option(
    "--number", "--execution",
    help="Pipeline execution number.",
    type=click.INT,
)
@click.option(
    "--step", "step_key",
    help="Step key.",
    type=click.STRING,
)
@click.option(
    "--yes",
    is_flag=True,
    default=False,
)
def resume_api(pipeline_name: str, number: int, step_key: str, yes: bool):
    """This command resumes selected pipeline step."""
    steps = get_paused_steps(pipeline_name=pipeline_name, execution_number=number, step_key=step_key)
    choices = []
    for step in steps:
        x = step.context.metadata.execution
        choices.append((format_step(step), (step.id, x.execution_number)))

    if len(choices) == 0:
        print_info("No matching idle pipeline execution found.")
        return
    elif len(choices) == 1:
        exec_id, number = choices[0][1]
    else:
        exec_id, number = prompt_choices("Choose step to resume", choices)

    if not yes:
        if not prompt_confirm(f"This will resume execution {pipeline_name} (#{number}). Are you sure?"):
            return

    resume_step_api(exec_id)
    print_info(f"{pipeline_name} (#{number}) has been resumed.")


@click.command()
def resume():
    """This command resumes current pipeline step.
    Only works in pipeline debugging mode."""
    if not in_debugging_context():
        print_warning("This command only works in pipeline debugging mode")
        return

    if prompt_confirm("This will terminate current debugging session and"
                      " resume pipeline step. Are you sure?", default=True):
        with open('/opt/vessl/.exit_code', 'w') as f:
            f.write("0")
        resume_step({"exit_code": "0"})
        print_info("Step resumed. Waiting for container to be terminated....")
        while True:
            time.sleep(1000)


@click.command
@click.option('--data', '-d', multiple=True)
def set_variable(data):
    """This command changes output variable for current step.
    Only works in pipeline debugging mode."""
    if not in_debugging_context():
        print_warning("This command only works in pipeline debugging mode")
        return

    vardict = {}
    for kvpair in data:
        try:
            k, v = kvpair.split(":")
        except ValueError:
            print_info(f"Invalid format: {kvpair}")
            return
        vardict[k] = v
    update_context_variables(vardict)


def in_debugging_context():
    try:
        with open("/opt/vessl/status", 'r') as f:
            return f.read() == 'idle'
    except FileNotFoundError:
        return False


if in_debugging_context():
    cli.add_command(resume)
    cli.add_command(set_variable)
else:
    cli.add_command(resume_api)



from typing import List, Any, Callable

import click
import inquirer

from vessl.util.echo import style_prompt


def prompt_confirm(text: str, default: bool = False) -> bool:
    """Prompt confirmation"""
    key = "question"
    inquiry = inquirer.Confirm(key, message=style_prompt(text), default=default)
    return inquirer.prompt([inquiry], raise_keyboard_interrupt=True).get(key)


def prompt_text(text: str, default: str = None) -> str:
    """Prompt text"""
    key = "question"
    inquiry = inquirer.Text(key, message=style_prompt(text), default=default)
    return inquirer.prompt([inquiry], raise_keyboard_interrupt=True).get(key)


def prompt_choices(text: str, choices: List[Any], default: Any = None) -> Any:
    """Prompt choices

    Args:
        choices (list): A list of choices to display, or a list of 2-tuples where
                        the first element is the choice to display and the second
                        element is return value.
    """
    key = "question"
    inquiry = inquirer.List(key, message=style_prompt(text), default=default, choices=choices)
    return inquirer.prompt([inquiry], raise_keyboard_interrupt=True).get(key)


def prompt_checkbox(text: str, choices: List[Any], default: Any = None) -> Any:
    key = "question"
    inquiry = inquirer.Checkbox(key, message=style_prompt(text), default=default, choices=choices)
    return inquirer.prompt([inquiry], raise_keyboard_interrupt=True).get(key)


def generic_prompter(text: str, type: click.ParamType = click.STRING, default=None) -> Callable:
    def prompter(ctx: click.Context, param: click.Parameter, value: str):
        return click.prompt(style_prompt(text), type=type, default=default)

    return prompter


def choices_prompter(text: str, choices: List[Any], default: Any = None) -> Callable:
    def prompter(ctx: click.Context, param: click.Parameter, value: str):
        return prompt_choices(style_prompt(text), choices, default)

    return prompter



import sys
from typing import Dict, List

import click
import inquirer

from vessl.openapi_client import OrmModelServiceGatewayTrafficSplitEntry as TrafficSplitEntry
from vessl.openapi_client import (
    ResponseKernelCluster,
    ResponseKernelImage,
    ResponseModelServiceGatewayInfo,
    ResponseModelServiceInfo,
    ResponseModelServiceRevision,
    ResponsePathOperation,
    ResponseSecret,
    ResponseVolumeSourceEntity,
    V1Autoscaling,
    V1EnvVar,
    V1ImportForm,
    V1Port,
    V1VariableValue,
)
from vessl.openapi_client.models import secret_generic_secret_upsert_api_input
from vessl import vessl_api
from vessl.cli._util import print_data, print_table_tabulate
from vessl.models.vessl_model import VesslModel
from vessl.secret import list_generic_secrets
from vessl.util.common import parse_time_to_ago
from vessl.util.constant import (
    VESSL_SERVICE_TENSORFLOW_BASE_IMAGE_TEMPLATE,
    VESSL_SERVICE_TORCH_BASE_IMAGE_TEMPLATE,
)
from vessl.util.echo import print_error, print_info
from vessl.util.exception import InvalidYAMLError
from vessl.util.prompt import prompt_text


def print_services(services: List[ResponseModelServiceInfo]):
    print(f"{len(services)} service(s) found.\n")
    services_sorted = sorted(services, key=lambda service: service.created_dt, reverse=True)
    data = []
    for service in services_sorted:
        kernel_cluster: ResponseKernelCluster = service.kernel_cluster
        status: str = service.status
        data.append({
            "Name": service.name,
            "Cluster": kernel_cluster.name,
            "Status": status,
            "Created At": f"{service.created_dt} ({parse_time_to_ago(service.created_dt)})"
        })
    print_table_tabulate(data)

def print_service(service: ResponseModelServiceInfo):
    data = {
        "Name": service.name,
        "Description": "(none)" if service.description=="" else service.description,
        "Created By": service.created_by.email,
        "Created At": service.created_dt,
        "Cluster": service.kernel_cluster.name,
        "Status": service.status,
        "Type": service.type,
        "Endpoint": service.gateway_config.endpoint if service.gateway_config.enabled else "(not enabled)",
    }
    print_data(data=data)

def print_revisions(revisions: List[ResponseModelServiceRevision], verbose: bool = False):
    revisions_sorted = sorted(revisions, key=lambda revision: revision.created_dt, reverse=True)
    if not verbose:
        revs = [
                {
                    "Revision Number": r.number,
                    "Message":r.message,
                    "Status": r.status,
                    "Created At": f"{r.created_dt} ({parse_time_to_ago(r.created_dt)})",
                }
                for r in revisions_sorted
            ]
        print_table_tabulate(revs)
        print("To view more details on each revision, use the --detail(-d) flag option.")
        print()
    else:
        for r in revisions_sorted:
            kernel_image: ResponseKernelImage = r.kernel_image
            autoscaler_config: V1Autoscaling = r.revision_spec.autoscaling
            ports: List[V1Port] = r.revision_spec.ports
            data = {
                f"Revision #{r.number}": {
                    "Status": r.status,
                    "Message": r.message,
                    "Available Replicas": r.available_replicas,
                    "Deployment Spec": {
                        "Image URL": kernel_image.image_url,
                        "Ports": [
                            {"Name": port.name, "Port": port.source_port, "Type": port.protocol}
                            for port in ports
                        ],
                    },
                    "Autoscaler Config": {
                        "Min Replicas": autoscaler_config.min,
                        "Max Replicas": autoscaler_config.max,
                        "Resource": autoscaler_config.metric,
                        "Target": autoscaler_config.target,
                    },
                }
            }
            print_data(data)
            print()

def print_revision(revision: ResponseModelServiceRevision, verbose: bool = False):
    if not verbose:
        data = {
            "Number": revision.number,
            "Status": revision.status,
            "Message": revision.message,
        }
    else:
        kernel_image: ResponseKernelImage = revision.kernel_image
        autoscaler_config: V1Autoscaling = revision.revision_spec.autoscaling
        ports: List[V1Port] = revision.revision_spec.ports
        data = {
            "Number": revision.number,
            "Status": revision.status,
            "Message": revision.message,
            "Available Replicas": revision.available_replicas,
            "Deployment Spec": {
                "Image URL": kernel_image.image_url,
                "Ports": [
                    {"Name": port.name, "Port": port.source_port, "Type": port.protocol}
                    for port in ports
                ],
            },
            "Autoscaler Config": {
                "Min Replicas": autoscaler_config.min,
                "Max Replicas": autoscaler_config.max,
                "Resource": autoscaler_config.metric,
                "Target": autoscaler_config.target,
            },
        }
    print_data(data)

def print_gateway(gateway: ResponseModelServiceGatewayInfo):
    def _prettify_traffic_split_entry(entry: TrafficSplitEntry) -> str:
        """
        Returns a pretty representation of given traffic split entry.

        Examples:
        - "########   80%:   3 (port 8000)"
        - "##         20%:  12 (port 3333)"
        """
        gauge_char = "#"
        gauge_width = (entry.traffic_weight + 9) // 10  # round up (e.g. 31%-40% -> 4)
        gauge = gauge_char * gauge_width

        return (
            f"{gauge: <10} {entry.traffic_weight: <3}%: {entry.revision_number: >3} "
        )

    print_data(
        {
            "Enabled": gateway.enabled,
            "Status": gateway.status,
            "Endpoint": gateway.endpoint or "(not set)",
            "Ingress Class": gateway.ingress_class or "(empty)",
            "Annotations": gateway.annotations or "(empty)",
            "Traffic Targets": (
                list(map(_prettify_traffic_split_entry, gateway.rules))
                if gateway.rules
                else "(empty)"
            ),
        }
    )


def validate_revision_yaml_obj(yaml_obj):
    """
    Client-side validation. We do not need a full validation here, as it will
    be done on the server side anyway. Only perform necessary validations to
    ensure that other client-side logics run safely.

    Raises:
        InvalidYAMLError: this YAML object for revision is invalid.
    """

    def _check_type(obj, type_: type):
        assert isinstance(obj, type_)
        return obj

    try:
        if "ports" in yaml_obj:
            ports: list = _check_type(yaml_obj["ports"], list)

            for port in ports:
                expose_type = _check_type(port["type"], str)
                assert expose_type in ["http", "tcp"]

                port_number = _check_type(port["port"], int)
                assert 1 <= port_number <= 65535

    except (KeyError, AssertionError) as e:
        raise InvalidYAMLError(message=str(e))


def list_http_ports(yaml_obj) -> List[int]:
    """
    Lists all HTTP expose ports of the revision.

    This function assumes the YAML object is valid (i.e. has valid port definitions.)
    """

    if "ports" not in yaml_obj:
        return []

    http_ports: List[int] = []

    try:
        ports: list = yaml_obj["ports"]
        for port in ports:
            expose_type: str = port["type"]
            if expose_type == "http":
                port_number: int = port["port"]
                http_ports.append(port_number)

    except (KeyError, AssertionError) as e:
        raise InvalidYAMLError(message=str(e))

    return http_ports

def select_secret():
    print_info("Select secret for authentication.")
    secrets: List[ResponseSecret] = list_generic_secrets()
    if len(secrets) == 0:
        print_info("No secrets registered.")
        if click.confirm("Do you want to create a new secret?", default=True):
            return create_secret()
        else:
            print_error("No secret selected.")
            sys.exit(1)
    else:
        secret_names = [x.credentials_name for x in secrets]
        secret_names.append("Create a new secret")
        secret_name = inquirer.prompt(
            [
                inquirer.List(
                    "secret",
                    message="Select from secrets",
                    choices=secret_names,
                )
            ],
            raise_keyboard_interrupt=True,
        ).get("secret")
        if secret_name == "Create a new secret":
            return create_secret()
        return secret_name

def fetch_variable_from_injects(key):
        input_method = inquirer.prompt(
            [
                inquirer.List(
                    "input_method",
                    message=f"{key}?",
                    choices=["Input as text", "Choose from VESSL secrets"],
                )
            ],
            raise_keyboard_interrupt=True,
        ).get("input_method")

        if input_method == "Input as text":
            return V1VariableValue(source="text", text=click.prompt(f"Enter value for {key}"))
        elif input_method == "Choose from VESSL secrets":
            return V1VariableValue(source="secret", secret=select_secret())

def create_secret() -> str:
    secret_name = prompt_text("Secret name")
    secret_value = prompt_text("Secret value")
    resp = vessl_api.secret_generic_secret_upsert_api(
        organization_name=vessl_api.organization.name,
        secret_generic_secret_upsert_api_input=secret_generic_secret_upsert_api_input.SecretGenericSecretUpsertAPIInput(
            secret_name=secret_name,
            value=secret_value,
        )
    )
    print_info(f"Secret {resp.credentials_name} created.")
    return resp.credentials_name


def _translate_vops(vops: List[ResponsePathOperation]) -> dict:
    def _parse_source_entity(e: ResponseVolumeSourceEntity, t: str) -> str:
        if t == "model":
            org = e.model.model_repository.organization.name
            model_repo_name = e.model.model_repository.name
            model_num = e.model.model.number
            return f"vessl-model://{org}/{model_repo_name}/{model_num}"
        if t == "dataset":
            org = e.dataset.organization.name
            dataset_name = e.dataset.name
            return f"vessl-dataset://{org}/{dataset_name}"
        if t == "artifact":
            name = e.artifact.persistent.name
            return f"vessl-artifact://{name}"
        if t == "git":
            provider = e.git.repository.provider
            owner = e.git.repository.owner
            repo = e.git.repository.repo
            if provider == "bitbucket":
                return f"git://bitbucket.org/{owner}/{repo}"
            if provider == "huggingface":
                return f"hf://huggingface.co/{owner}/{repo}"
            else:
                return f"git://{provider}.com/{owner}/{repo}"
        if t == "s3":
            bucket = e.s3.bucket
            prefix = e.s3.prefix
            return f"s3://{bucket}/{prefix}"
        if t == "gs":
            bucket = e.gs.bucket
            prefix = e.gs.prefix
            return f"gs://{bucket}/{prefix}"
        if t == "hostpath":
            path = e.hostpath.path
            return f"hostpath://{path}"
        if t == "nfs":
            server = e.nfs.server
            path = e.nfs.path
            return f"nfs://{server}/{path}"
        if t == "cifs":
            return None
        if t == "googledisk":
            return None
        else:
            return None

    res = {
        "import": {},
        "mount": {},
    }
    for item in vops:
        if item._import:
            source_entity = item._import.source_entity
            t = source_entity.source_entity_type
            signature = _parse_source_entity(source_entity, t)
            res["import"][item.path] = signature
        if item.mount:
            source_entity = item.mount.source_entity
            t = source_entity.source_entity_type
            signature = _parse_source_entity(source_entity, t)
            res["mount"][item.path] = signature
    if not res["import"]:
        del res["import"]
    if not res["mount"]:
        del res["mount"]
    return res

def _translate_envvars(envvars: Dict[str,V1EnvVar]) -> dict:
    res = {}
    for k, v in envvars.items():
        if v.default_value.source == "secret":
            res[k] = {
                "secret": v.default_value.secret,
                "source": v.default_value.source,
            }
        if v.default_value.source == "text":
            res[k] = {
                "text": v.default_value.text,
                "source": v.default_value.source,
            }
    return res

def get_runner_type(model_runner_type: str) -> str:
    if model_runner_type == "vessl" or model_runner_type == "hf-transformers" or model_runner_type == "torch":
        return "vessl"
    elif model_runner_type == "bento":
        return "bento"
    else:
        raise ValueError(f"Unsupported model runner type: {model_runner_type}")
    
def sanitize(yaml_obj):
    if isinstance(yaml_obj, dict):
        cleaned = dict((k, sanitize(v)) for k, v in yaml_obj.items() if v is not None)
        return cleaned
    elif isinstance(yaml_obj, list):
        cleaned = [sanitize(v) for v in yaml_obj if v is not None]
        return cleaned
    else:
        return yaml_obj

def sanitize_model_form(form: V1ImportForm):
    return f"vessl-model://{form.model.organization_name}/{form.model.model_repository_name}/{form.model.model_number}"

def generate_image_name(
    vesslmodel: VesslModel,
):
    if vesslmodel.framework_type == "torch":
        return VESSL_SERVICE_TORCH_BASE_IMAGE_TEMPLATE.format(
            **{
                "pytorch_version": vesslmodel.pytorch_version,
                "cuda_version": vesslmodel.cuda_version,
            }
        )
    elif vesslmodel.framework_type == "tensorflow":
        return VESSL_SERVICE_TENSORFLOW_BASE_IMAGE_TEMPLATE.format(
            **{
                "tensorflow_version": vesslmodel.tensorflow_version,
                "cuda_version": vesslmodel.cuda_version,
            }
        )
    raise ValueError("Either pytorch_version or tensorflow_version must be provided")


import abc
import os
import pickle
import subprocess
import sys
import time
from importlib import import_module
from io import BufferedReader
from typing import Any, Callable, Dict, Generic, List, Tuple, TypeVar, Union, Optional

import yaml
from pydantic import BaseModel

from vessl import vessl_api
from vessl.util.fmt import format_url
from vessl.util.endpoint import Endpoint
from vessl.model import (
    create_model,
    download_model_volume_file,
    list_model_volume_files,
    read_model,
    upload_model_volume_file,
)
from vessl.organization import _get_organization_name
from vessl.util import logger
from vessl.util.common import get_module
from vessl.util.exception import (
    InvalidParamsError,
    VesslApiException,
    VesslRuntimeException,
)
from vessl.util.constant import LOGO, WEB_HOST, colors

from vessl.serving import (
    get_recent_rollout,
    abort_in_progress_rollout_by_name,
    create_revision_from_yaml,
    create_active_revision_replacement_rollout,
    launch_revision,
    _wait_for_revision_to_launch,
    _wait_for_gateway_enabled,
    _read_service,
)

ModelType = TypeVar("ModelType")
InputDataType = TypeVar("InputDataType", bound=BaseModel)
ModelInputDataType = TypeVar("ModelInputDataType")
ModelOutputDataType = TypeVar("ModelOutputDataType")
OutputDataType = TypeVar("OutputDataType", bound=BaseModel)

PROTOCOL_VERSION = 4
USER_DEFINED = "user-defined"
MANIFEST_FILENAME = "vessl.manifest.yaml"
RUNNER_FILENAME = "vessl.runner.pkl"
DEFAULT_MODEL_FILENAME = "vessl.model.pkl"


class RunnerBase(abc.ABC, Generic[InputDataType, OutputDataType]):
    """Base class for model registering.

    This base class introduces 5 static methods as followings:
    - `predict`: Make prediction with given data and model. This method must be overridden. The
      data is given from the result of `preprocess_data`, and the return value of this method
      will be passed to `postprocess_data` before serving.
    - `save_model`: Save the model into a file. Return value of this method will be given to the
      `load_model` method on model loading. If this method is overriden, `load_model` must be
      overriden as well.
    - `load_model`: Load the model from a file.
    - `preprocess_data`: Preprocess the data before prediction. It converts the API input data to
      the model input data.
    - `postprocess_data`: Postprocess the data after prediction. It converts the model output data
      to the API output data.

    Check each method's docstring for more information.
    """

    @staticmethod
    def save_model(model: ModelType) -> Dict[str, str]:
        """Save the given model instance into file.

        Return value of this method will be given to first argument of `load_model` on model loading.

        Args:
            model(model_instance): Model instance to save.

        Returns:
            (dict) Data that will be passed to `load_model` on model loading.
                Must be a dictionary with key and value both string.
        """
        pass

    @staticmethod
    def load_model(props: Union[Dict[str, str], None], artifacts: Dict[str, str]) -> ModelType:
        """Load the model instance from file.

        `props` is given from the return value of `save_model`, and `artifacts` is
        given from the `register_model` method.

        If the `save_model` is not overriden, `props` will be None

        Args:
            props(dict | None): Data that was returned by `save_model`. If `save_model` is
                not overriden, this will be None.
            artifacts(dict): Data that is given by `register_model` function.

        Returns:
            Model instance.
        """
        pass

    @staticmethod
    def preprocess_data(data: InputDataType) -> ModelInputDataType:
        """Preprocess the given data.

        The data processed by this method will be given to the model.

        Args:
            data: Data to be preprocessed.

        Returns:
            Preprocessed data that will be given to the model.
        """
        pass

    @staticmethod
    def postprocess_data(data: ModelOutputDataType) -> OutputDataType:
        """Postprocess the given data.

        The data processed by this method will be given to the user.

        Args:
            data: Data to be postprocessed.

        Returns:
            Postprocessed data that will be given to the user.
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def predict(model: ModelType, data: ModelInputDataType) -> ModelOutputDataType:
        """Make prediction with given data and model.

        Args:
            model(model_instance): Model instance.
            data: Data to be predicted.

        Returns:
            Prediction result.
        """
        raise NotImplementedError()


class ModelConfig(BaseModel):
    save_type: Optional[str] = None
    save_path: Optional[str] = None
    save_output: Optional[Dict[str, str]] = None


class RunnerConfig(BaseModel):
    save_type: str
    save_path: str


class ManifestSchema(BaseModel):
    version: str
    model: ModelConfig
    runner: RunnerConfig
    use_default: List[str]
    requirements: List[str]
    artifacts: Optional[Dict[str, str]] = {}


def _save_object(save_type: Union[str, None], obj: Any, path: str) -> str:
    """Save the object into `path`

    If `save_type` is not `None`, use the given serialization method. It will not fallback on failure.
    Otherwise, use various methods to serialize the object as best as possible.
    After serialization, save it into `path`.

    Args:
        save_type (str | None): The serialization method.
        obj (Any): The object to serialize.
        path (str): The path to save the object.

    Returns:
        (str) The save type finally applied to the object.
    """
    # Known save types
    if save_type == "torch":
        import cloudpickle
        import torch

        with open(path, "wb") as f:
            torch.save(obj, f, pickle_module=cloudpickle, pickle_protocol=PROTOCOL_VERSION)
        return "torch"
    if save_type is not None:
        # save_type given, but we don't know how to handle it
        raise InvalidParamsError("Unknown save type: {}".format(save_type))

    # Fallback saving methods for None save_type
    with open(path, "wb") as f:
        # Try cloudpickle first
        try:
            import cloudpickle

            cloudpickle.dump(obj, f, protocol=PROTOCOL_VERSION)
            return "cloudpickle"
        except pickle.PicklingError:
            pass

        # Try joblib
        try:
            joblib = get_module("joblib")
            joblib.dump(obj, f)
            return "joblib"
        except pickle.PicklingError:
            pass

        # Last fallback. Use pickle module
        try:
            pickle.dump(obj, f, protocol=PROTOCOL_VERSION)
            return "pickle"
        except pickle.PicklingError:
            pass
    # Failed to save. Raise.
    raise VesslRuntimeException("Could not save the object")


def _try_load(loader: Callable, br: BufferedReader) -> Any:
    try:
        return loader(br)
    except Exception as e:
        raise VesslRuntimeException(
            f"Could not load the object. Did you use a different Python version to register the model?\nError: {e}"
        )


def _load_object(save_type: str, path: str) -> Any:
    """Load the object from given path

    It uses the given save_type to find a correct way to deserialize the object.

    Args:
        save_type (str): The serialization method used to save the object.
        path (str): The path to load the object.

    Returns:
        (Any) The loaded object.
    """
    if save_type == "torch":
        with open(path, "rb") as f:
            torch = get_module(
                "torch", required='torch package is required. Run "pip install torch".'
            )
            return _try_load(torch.load, f)

    with open(path, "rb") as f:
        if save_type == "cloudpickle":
            import cloudpickle

            return _try_load(cloudpickle.load, f)
        if save_type == "joblib":
            joblib = get_module(
                "joblib",
                required='joblib package is required. Run "pip install joblib".',
            )
            return _try_load(joblib.load, f)
        if save_type == "pickle":
            return _try_load(pickle.load, f)
    raise InvalidParamsError("Unknown save type")


def _upload_artifacts(
    artifacts: Dict[str, str],
    *,
    repository_name: str,
    model_number: int,
    organization_name: str,
) -> None:
    """Upload given artifacts to the model volume.

    Args:
        artifacts(dict): Artifacts to be uploaded. Key is the path to artifact in local filesystem, and value is the
            path in the model volume. Only trailing asterisk(*) is allowed for glob pattern.
        repository_name(str): Name of the repository.
        model_number(int): Number of the model.
        organization_name(str): Name of the organization.

    Example:
        >>> _upload_artifacts({
        >>>     "./output/*": "*",
        >>>     "./sample.data": "sample.data",
        >>> }, repository_name="my-repository", model_number=1, organization_name="my-organization")
    """
    # Input Validation
    for local_path, remote_path in artifacts.items():
        local_dir, local_base = os.path.dirname(local_path), os.path.basename(local_path)
        remote_dir, remote_base = os.path.dirname(remote_path), os.path.basename(remote_path)
        if "*" in local_dir or "*" in remote_dir:
            raise InvalidParamsError("Only trailing asterisk(*) is allowed for glob pattern")
        if ("*" in local_base and local_base != "*") or ("*" in remote_base and remote_base != "*"):
            raise InvalidParamsError("Only trailing asterisk(*) is allowed for glob pattern")
        if (local_base == "*") != (remote_base == "*"):
            # Both should either or neither be glob pattern
            raise InvalidParamsError(
                "Both local and remote path should either or neither be glob pattern"
            )

    if len(artifacts.keys()) != 0:
        print("Artifacts will be uploaded as follows:")
        for l, r in artifacts.items():
            print("  {} -> {}".format(l, r))

    for local_path, remote_path in artifacts.items():
        local_dir, local_base = os.path.dirname(local_path), os.path.basename(local_path)
        remote_dir, remote_base = os.path.dirname(remote_path), os.path.basename(remote_path)
        if local_base == "*":
            # Glob pattern
            for root, _, files in os.walk(local_dir):
                for filename in files:
                    local_filepath = os.path.join(root, filename)
                    upload_model_volume_file(
                        repository_name=repository_name,
                        model_number=model_number,
                        organization_name=organization_name,
                        source_path=local_filepath,
                        dest_path=local_filepath.replace(local_dir, remote_dir),
                    )
        else:
            # Upload single file
            upload_model_volume_file(
                repository_name=repository_name,
                model_number=model_number,
                organization_name=organization_name,
                source_path=local_path,
                dest_path=remote_path,
            )


# Related to model registration. Will be used as Python SDK, not CLI.
def _register_serving_method(
    organization_name: str,
    repository_name: str,
    model_number: Union[int, None],
    model_type: Union[str, None],
    model_instance: Union[ModelType, None],
    runner_cls: RunnerBase,
    requirements: List[str],
    artifacts: Dict[str, str],
):
    # Argument validation
    # Valid comb of (model_number, model_type, model_instance) (0 means None)
    # - (0, 0, 1): Create new model from given model_instance and register. Default fallback saver will be used since model_type is not given.
    # - (0, 1, 1): Create new model from given model_instance and register. model_type's saver will be used.
    # - (1, 0, 0): Register existing model. It will only add runner/manifest file to make model servable.
    # - (1, 0, 1): Overwrite existing model. Given model_instance will be serialized and saved to given model number. Default fallback saver will be used since model_type is not given.
    # - (1, 1, 1): Overwrite existing model. Given model_instance will be serialized and saved to given model number. model_type's saver will be used.
    if model_number is None and model_instance is None:  # Exclude (0, x, 0)
        raise InvalidParamsError("`model_instance` must be given to create a new model")
    if model_instance is None and model_type is not None:  # Exclude (x, 1, 0)
        raise InvalidParamsError("`model_instance` must be given if `model_type` is not None")
    for key in artifacts:
        if key in [RUNNER_FILENAME, MANIFEST_FILENAME, DEFAULT_MODEL_FILENAME]:
            raise InvalidParamsError("`{}` is a reserved key of artifacts".format(key))

    if model_number is not None:
        # Check model existence
        try:
            read_model(
                repository_name=repository_name,
                model_number=model_number,
                organization_name=organization_name,
            )
        except VesslApiException:
            raise InvalidParamsError("Failed to read requested model")
    else:
        # Create new model
        logger.info("`model_number` is None. New model will be created.")
        res = create_model(
            repository_name=repository_name,
            organization_name=organization_name,
        )
        model_number = res.number

    # Runner class validation
    runner_cls()  # Emits error if abstract method is not overriden
    if model_instance is None:
        # Tries to add serving method only.
        if runner_cls.load_model == RunnerBase.load_model:
            raise InvalidParamsError(
                "`load_model` method must be overriden if model_instance is not given"
            )
    else:
        if (
            runner_cls.save_model != RunnerBase.save_model
            and runner_cls.load_model == RunnerBase.load_model
        ):
            raise InvalidParamsError(
                "`load_model` must be implemented if `save_model` is implemented"
            )
    # Validation done.

    # Business logic from here
    # Save the model
    model_save_type, model_save_path, model_save_output = USER_DEFINED, None, None
    if model_instance is not None:
        if runner_cls.save_model != RunnerBase.save_model:
            # Use the custom `save_model` if defined
            model_save_type = USER_DEFINED
            model_save_output = runner_cls.save_model(model_instance)
        elif runner_cls.load_model != RunnerBase.load_model:
            # `save_model` is not overriden but `load_model` is overridden.
            # We can assume user saves the model in other way. Skip saving
            model_save_type = USER_DEFINED
        else:
            # Use the default saver
            model_save_type = _save_object(model_type, model_instance, DEFAULT_MODEL_FILENAME)
            model_save_path = DEFAULT_MODEL_FILENAME

    # Save the runner
    runner_save_type, runner_save_path = "", ""
    runner_save_type = _save_object(None, runner_cls, RUNNER_FILENAME)
    runner_save_path = RUNNER_FILENAME

    # Mark the methods that are not overriden by user
    use_default = []
    if runner_cls.load_model == RunnerBase.load_model:
        use_default.append("load_model")
    if runner_cls.preprocess_data == RunnerBase.preprocess_data:
        use_default.append("preprocess_data")
    if runner_cls.postprocess_data == RunnerBase.postprocess_data:
        use_default.append("postprocess_data")

    # Save the manifest
    manifest = {
        "version": "v1",
        "model": {
            "save_type": model_save_type,
            "save_path": model_save_path,
            "save_output": model_save_output,
        },
        "runner": {
            "save_type": runner_save_type,
            "save_path": runner_save_path,
        },
        "use_default": use_default,
        "requirements": requirements,
        "artifacts": artifacts,
    }
    ManifestSchema.model_validate(manifest)
    with open(MANIFEST_FILENAME, "w") as f:
        yaml.dump(manifest, f)

    # Upload artifacts
    upload = artifacts.copy()
    upload[MANIFEST_FILENAME] = MANIFEST_FILENAME
    upload[runner_save_path] = runner_save_path
    if model_save_path is not None:
        upload[model_save_path] = model_save_path
    _upload_artifacts(
        upload,
        repository_name=repository_name,
        model_number=model_number,
        organization_name=organization_name,
    )

    print(
        f"Successfully registered model: {format_url(Endpoint.model.format(organization_name, repository_name, model_number))}"
    )


def register_model(
    repository_name: str,
    model_number: Union[int, None],
    runner_cls: RunnerBase,
    model_instance: Union[ModelType, None] = None,
    requirements: List[str] = None,
    artifacts: Dict[str, str] = None,
    **kwargs,
):
    """Register the given model for serving. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        repository_name(str): Model repository name.
        model_number(int | None): Model number. If None, new model will be
            created. In such case, `model_instance` must be given.
        runner_cls(RunnerBase): Runner class that includes code for serving.
        model_instance(ModelType | None): Model instance. If None, `runner_cls`
            must override `load_model` method. Defaults to None.
        requirements(List[str]): Python requirements for the model. Defaults to
            [].
        artifacts(Dict[str, str]): Artifacts to be uploaded. Key is the path to
            artifact in local filesystem, and value is the path in the model
            volume. Only trailing asterisk(*) is allowed for glob pattern.
            Defaults to {}.

    Example:
        ```python
        register_model(
            repository_name="my-model",
            model_number=1,
            runner_cls=MyRunner,
            model_instance=model_instance,
            requirements=["torch", "torchvision"],
            artifacts={"model.pt": "model.pt", "checkpoints/*": "checkpoints/*"},
        )
        ```
    """
    organization_name = _get_organization_name(**kwargs)
    if requirements is None:
        requirements = []
    if artifacts is None:
        artifacts = {}
    _register_serving_method(
        organization_name=organization_name,
        repository_name=repository_name,
        model_number=model_number,
        model_type=None,
        model_instance=model_instance,
        runner_cls=runner_cls,
        requirements=requirements,
        artifacts=artifacts,
    )


def register_torch_model(
    repository_name: str,
    model_number: Union[int, None],
    model_instance: ModelType,
    preprocess_data=None,
    postprocess_data=None,
    requirements: List[str] = None,
    **kwargs,
):
    """Register the given torch model instance for model serving. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        repository_name(str): Model repository name.
        model_number(int | None): Model number. If None, new model will be
            created.
        model_instance(model_instance): Torch model instance.
        preprocess_data(callable): Function that will preprocess data.
            Defaults to identity function.
        postprocess_data(callable): Function that will postprocess data.
            Defaults to identity function.
        requirements(list): List of requirements. Defaults to [].

    Example:
        ```python
        vessl.register_model(
            repository_name="my-model",
            model_number=1,
            model_instance=model_instance,
            preprocess_data=lambda x: int(x),
            postprocess_data=lambda x: {"prediction": x},
            requirements=["torch", "torchvision"],
        )
        ```
    """
    # Check if the model is torch
    torch_nn = get_module(
        "torch.nn", required='torch package is required. Run "pip install torch".'
    )
    if not isinstance(model_instance, torch_nn.Module):
        raise InvalidParamsError("model_instance must be a torch.nn.Module.")

    organization_name = _get_organization_name(**kwargs)
    if requirements is None:
        requirements = []

    class PyTorchRunner(RunnerBase):
        @staticmethod
        def predict(model: ModelType, data: ModelInputDataType) -> ModelOutputDataType:
            return model(data)

    if preprocess_data is not None:
        PyTorchRunner.preprocess_data = preprocess_data
    if postprocess_data is not None:
        PyTorchRunner.postprocess_data = postprocess_data
    _register_serving_method(
        organization_name=organization_name,
        repository_name=repository_name,
        model_number=model_number,
        model_type="torch",
        model_instance=model_instance,
        runner_cls=PyTorchRunner,
        requirements=requirements + ["torch", "torchvision"],
        artifacts={},
    )


# Related to model serving. Will be used by CLI.
class _ModelServer(Generic[InputDataType, OutputDataType]):
    """Model server class."""

    def __init__(self, model: ModelType, preprocess_data, predict, postprocess_data):
        self.model = model
        self.preprocess_data = preprocess_data
        self.predict = predict
        self.postprocess_data = postprocess_data

    def run(self, data: InputDataType) -> OutputDataType:
        processed = self.preprocess_data(data)
        result = self.predict(self.model, processed)
        return self.postprocess_data(result)


def _load_model(install_reqs: bool = False) -> _ModelServer:
    """Load model from file. This function assumes that cwd is the root of model volume.

    Args:
        install_reqs(bool): Whether to install requirements. Default is False.

    Returns:
        (ModelServer) Model server.
    """
    with open(MANIFEST_FILENAME, "r") as f:
        try:
            raw_manifest = yaml.load(f, Loader=yaml.FullLoader)
        except yaml.YAMLError:
            raise VesslRuntimeException("Invalid manifest file. Please register the model again.")

    manifest = ManifestSchema.model_validate(raw_manifest)
    # Check for requirements
    if install_reqs:
        reqs = manifest.requirements
        if reqs:
            logger.info("Installing requirements:", reqs)
            subprocess.check_call(["pip", "install", *reqs])  # Support for conda?

    # Load runner
    runner_cls = _load_object(manifest.runner.save_type, manifest.runner.save_path)
    # Load model
    if "load_model" in manifest.use_default:
        model = _load_object(manifest.model.save_type, manifest.model.save_path)
    else:  # Use custom loader
        model = runner_cls.load_model(manifest.model.save_output, manifest.artifacts)
    # Load methods
    preprocess_data = (
        (lambda x: x)
        if "preprocess_data" in manifest.use_default
        else runner_cls.preprocess_data
    )
    postprocess_data = (
        (lambda x: x)
        if "postprocess_data" in manifest.use_default
        else runner_cls.postprocess_data
    )
    model_server = _ModelServer(model, preprocess_data, runner_cls.predict, postprocess_data)
    return model_server


def _load_model_v2(
    filename: str, classname: str
) -> Tuple[_ModelServer, BaseModel, BaseModel, bool]:
    sys.path.append(os.getcwd())
    if not os.path.exists(filename) and not os.path.exists(f"{filename}.py"):
        raise FileNotFoundError(f"{filename} not found")

    module_name = filename.replace(".py", "")
    runner_cls = getattr(import_module(module_name), classname)
    runner = runner_cls()
    if not isinstance(runner, RunnerBase):
        raise ValueError(f"{classname} must be a subclass of RunnerBase")

    try:
        [input_type, output_type] = runner.__orig_bases__[0].__args__
    except:
        input_type, output_type = None, None

    try:
        if issubclass(input_type, BaseModel) or issubclass(output_type, BaseModel):
            pass_type_param = True
    except:
        pass_type_param = False

    model = runner.load_model(props=None, artifacts={})

    model_server = _ModelServer[input_type, output_type](
        model, runner.preprocess_data, runner.predict, runner.postprocess_data
    )

    return (model_server, input_type, output_type, pass_type_param)


def _run_model_server(
    model: _ModelServer,
    remote: bool,
    port: int = 8000,
    api_name: str = None,
    inputType: TypeVar = InputDataType,
    outputType: TypeVar = OutputDataType,
):
    raise NotImplementedError("This method is deprecated.")


def serve_model(
    repository_name: str,
    model_number: int,
    install_reqs: bool = False,
    remote: bool = False,
    **kwargs,
):
    """Execute the server which serves the given model. If you want to override the default organization, then pass `organization_name` as `**kwargs`.

    This method will create `.vessl-serving` under current directory to save the artifacts in local.

    Args:
        repository_name(str): Model repository name.
        model_number(int): Model number.
        install_reqs(bool): Whether to install requirements. Defaults to False.
    """
    # Validation
    organization_name = _get_organization_name(**kwargs)
    model = read_model(repository_name=repository_name, model_number=model_number, **kwargs)
    if not model.is_servable:
        raise InvalidParamsError("Model is not servable.")

    if not remote:
        print("Downloading model artifacts for serving...")
        # Create temporal directory if not exists
        local_path_root = f".vessl-serving/{organization_name}/{repository_name}/{model_number}"
        os.makedirs(local_path_root, exist_ok=True)

        # Get the list of files to be downloaded. Only the outdated files or the files that are not in the local will be downloaded.
        volume_files = list_model_volume_files(
            repository_name=repository_name, model_number=model_number, recursive=True
        )
        local_status_filename = (
            ".vessl.localstatus.yaml"  # File that keeps track of the local status of the files
        )
        local_status = {  # Structure of the local status file
            "hash": {},
        }
        download_list = []
        if os.path.exists(os.path.join(local_path_root, local_status_filename)):
            # Local status file exists -> do not download the files that are already up-to-date.
            with open(os.path.join(local_path_root, local_status_filename), "r") as f:
                local_status.update(yaml.load(f, Loader=yaml.FullLoader))
            for volume_file in volume_files:
                if volume_file.is_dir:
                    continue  # Skip directory
                if volume_file.hash == local_status.get("hash", {}).get(volume_file.path, None):
                    continue  # Skip if hash is up to date
                download_list.append(volume_file)
        else:
            # Local status file does not exist -> download all files.
            for volume_file in volume_files:
                if volume_file.is_dir:
                    continue
                download_list.append(volume_file)
        # Download list retrieved.

        # Download files
        os.chdir(local_path_root)  # Make root of the model volume as cwd
        for volume_file in download_list:
            download_model_volume_file(
                repository_name=repository_name,
                model_number=model_number,
                source_path=volume_file.path,
                dest_path=volume_file.path,
            )
            # Update hash so that we don't download the same file again.
            local_status["hash"][volume_file.path] = volume_file.hash
        with open(local_status_filename, "w") as f:
            yaml.dump(local_status, f)
        print("Artifacts for model serving are up to date.")

    # Serve model
    print("Loading model from artifacts...")
    model = _load_model(install_reqs)
    print("Booting local server for serving...")
    _run_model_server(model, remote)

def launch_service_revision_from_yaml(
    service_name: str,
    yaml_path: str,
    launch: bool = True,
    set_current_active: bool = True,
    ignore_rollout: bool = True,
    **kwargs,
):
    """Simply launch the service via yaml file. If you want to override the default organization, then pass `organization_name` as `**kwargs`.

    Args:
        service_name(str): Service Name.
        yaml_path(str): Yaml file path.
        launch(bool): If it's False, it just create a revision and do not launch the service. Default to True.
        set_current_active(bool): If it's True, current revision will get 100% traffic_weight. If it's False, it just create a revision. Default to True.
        ignore_rollout(bool): Whether to ignore current running rollout. Default to True.
    """
    assert launch == True or set_current_active == True

    organization_name = _get_organization_name(**kwargs)

    with open(yaml_path, "r") as f:
        yaml_str = f.read()
        try:
            yaml_body = yaml.load(f, Loader=yaml.FullLoader)
        except yaml.YAMLError as e:
            print(f"Error: invalid YAML\n{e}")
            sys.exit(1)

    ## Check if the service is already rolling out
    ongoing_rollout_exists = False
    try:
        recent_rollout = get_recent_rollout(service_name)
        if recent_rollout and recent_rollout.status == "rolling_out":
            ongoing_rollout_exists = True
            print(f"Error: the service {service_name} is currently rolling out.")
            if not ignore_rollout:
                print("Use ignore_rollout option to abort the existing rollout.")
                sys.exit(1)
    except VesslApiException as e:
        print("No existing rollout found.")
        pass

    ## Abort the existing rollout if ignore_rollout
    if ongoing_rollout_exists and ignore_rollout:
        if abort_in_progress_rollout_by_name(service_name):
            print("Waiting for the existing rollout to be aborted...")
            time.sleep(30)

    ## create revision
    try:
        revision = create_revision_from_yaml(
            organization=organization_name, yaml_body=yaml_str
        )
    except VesslApiException as e:
        if e.status == 400:
            print("Error: failed to create revision (invalid parameters).")
            print(e.message)
        else:
            print("Error: failed to create revision. (internal error).")
        sys.exit(1)

    ## launch or set-active
    if set_current_active:
        launch = True
    try:
        if set_current_active:
            create_active_revision_replacement_rollout(
                organization=organization_name,
                model_service_name=revision.model_service_name,
                desired_active_revisions_to_weight_map={revision.number: 100},
            )
        elif launch:
            launch_revision(
                organization=organization_name,
                service_name=revision.model_service_name,
                revision_number=revision.number,
            )
        print("Successfully triggered revision launch.")
        _wait_for_revision_to_launch(service_name=service_name, revision_number=revision.number, print_output=True)
        revision = vessl_api.model_service_revision_read_api(
            organization_name=vessl_api.organization.name,
            model_service_name=service_name,
            revision_number=revision.number,
        )
        print(revision)
        link = f"{WEB_HOST}/{organization_name}/services/{service_name}"
        print(f"Check your Service at: {link}")

    except VesslApiException as e:
        print("Error: failed to launch revision.")
        print(e.message)
        sys.exit(1)

    ## Wait for the gateway to be enabled
    if set_current_active:
        gateway = _read_service(service_name=service_name).gateway_config
        _wait_for_gateway_enabled(gateway=gateway, service_name=revision.model_service_name, print_output=True)

        print("Endpoint is enabled.")
        gateway = _read_service(service_name=service_name).gateway_config
        print(gateway)
        print(f"You can test your service via {gateway.endpoint}")



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class V1ImportForm(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'artifact': 'V1ImportableArtifactRefForm',
        'dataset': 'V1DatasetRefForm',
        'git': 'V1GitRefForm',
        'gs': 'V1GSRefForm',
        'hf': 'V1HuggingfaceRefForm',
        'model': 'V1ModelRefForm',
        'ref_type': 'str',
        's3': 'V1S3RefForm',
        'volume': 'V1VolumeV2ImportForm'
    }

    attribute_map = {
        'artifact': 'artifact',
        'dataset': 'dataset',
        'git': 'git',
        'gs': 'gs',
        'hf': 'hf',
        'model': 'model',
        'ref_type': 'ref_type',
        's3': 's3',
        'volume': 'volume'
    }

    def __init__(self, artifact=None, dataset=None, git=None, gs=None, hf=None, model=None, ref_type=None, s3=None, volume=None, local_vars_configuration=None):  # noqa: E501
        """V1ImportForm - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._artifact = None
        self._dataset = None
        self._git = None
        self._gs = None
        self._hf = None
        self._model = None
        self._ref_type = None
        self._s3 = None
        self._volume = None
        self.discriminator = None

        if artifact is not None:
            self.artifact = artifact
        if dataset is not None:
            self.dataset = dataset
        if git is not None:
            self.git = git
        if gs is not None:
            self.gs = gs
        if hf is not None:
            self.hf = hf
        if model is not None:
            self.model = model
        if ref_type is not None:
            self.ref_type = ref_type
        if s3 is not None:
            self.s3 = s3
        if volume is not None:
            self.volume = volume

    @property
    def artifact(self):
        """Gets the artifact of this V1ImportForm.  # noqa: E501


        :return: The artifact of this V1ImportForm.  # noqa: E501
        :rtype: V1ImportableArtifactRefForm
        """
        return self._artifact

    @artifact.setter
    def artifact(self, artifact):
        """Sets the artifact of this V1ImportForm.


        :param artifact: The artifact of this V1ImportForm.  # noqa: E501
        :type artifact: V1ImportableArtifactRefForm
        """

        self._artifact = artifact

    @property
    def dataset(self):
        """Gets the dataset of this V1ImportForm.  # noqa: E501


        :return: The dataset of this V1ImportForm.  # noqa: E501
        :rtype: V1DatasetRefForm
        """
        return self._dataset

    @dataset.setter
    def dataset(self, dataset):
        """Sets the dataset of this V1ImportForm.


        :param dataset: The dataset of this V1ImportForm.  # noqa: E501
        :type dataset: V1DatasetRefForm
        """

        self._dataset = dataset

    @property
    def git(self):
        """Gets the git of this V1ImportForm.  # noqa: E501


        :return: The git of this V1ImportForm.  # noqa: E501
        :rtype: V1GitRefForm
        """
        return self._git

    @git.setter
    def git(self, git):
        """Sets the git of this V1ImportForm.


        :param git: The git of this V1ImportForm.  # noqa: E501
        :type git: V1GitRefForm
        """

        self._git = git

    @property
    def gs(self):
        """Gets the gs of this V1ImportForm.  # noqa: E501


        :return: The gs of this V1ImportForm.  # noqa: E501
        :rtype: V1GSRefForm
        """
        return self._gs

    @gs.setter
    def gs(self, gs):
        """Sets the gs of this V1ImportForm.


        :param gs: The gs of this V1ImportForm.  # noqa: E501
        :type gs: V1GSRefForm
        """

        self._gs = gs

    @property
    def hf(self):
        """Gets the hf of this V1ImportForm.  # noqa: E501


        :return: The hf of this V1ImportForm.  # noqa: E501
        :rtype: V1HuggingfaceRefForm
        """
        return self._hf

    @hf.setter
    def hf(self, hf):
        """Sets the hf of this V1ImportForm.


        :param hf: The hf of this V1ImportForm.  # noqa: E501
        :type hf: V1HuggingfaceRefForm
        """

        self._hf = hf

    @property
    def model(self):
        """Gets the model of this V1ImportForm.  # noqa: E501


        :return: The model of this V1ImportForm.  # noqa: E501
        :rtype: V1ModelRefForm
        """
        return self._model

    @model.setter
    def model(self, model):
        """Sets the model of this V1ImportForm.


        :param model: The model of this V1ImportForm.  # noqa: E501
        :type model: V1ModelRefForm
        """

        self._model = model

    @property
    def ref_type(self):
        """Gets the ref_type of this V1ImportForm.  # noqa: E501


        :return: The ref_type of this V1ImportForm.  # noqa: E501
        :rtype: str
        """
        return self._ref_type

    @ref_type.setter
    def ref_type(self, ref_type):
        """Sets the ref_type of this V1ImportForm.


        :param ref_type: The ref_type of this V1ImportForm.  # noqa: E501
        :type ref_type: str
        """
        allowed_values = ["git", "hf", "dataset", "model", "s3", "gs", "artifact", "volume-v2"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and ref_type not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `ref_type` ({0}), must be one of {1}"  # noqa: E501
                .format(ref_type, allowed_values)
            )

        self._ref_type = ref_type

    @property
    def s3(self):
        """Gets the s3 of this V1ImportForm.  # noqa: E501


        :return: The s3 of this V1ImportForm.  # noqa: E501
        :rtype: V1S3RefForm
        """
        return self._s3

    @s3.setter
    def s3(self, s3):
        """Sets the s3 of this V1ImportForm.


        :param s3: The s3 of this V1ImportForm.  # noqa: E501
        :type s3: V1S3RefForm
        """

        self._s3 = s3

    @property
    def volume(self):
        """Gets the volume of this V1ImportForm.  # noqa: E501


        :return: The volume of this V1ImportForm.  # noqa: E501
        :rtype: V1VolumeV2ImportForm
        """
        return self._volume

    @volume.setter
    def volume(self, volume):
        """Sets the volume of this V1ImportForm.


        :param volume: The volume of this V1ImportForm.  # noqa: E501
        :type volume: V1VolumeV2ImportForm
        """

        self._volume = volume

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1ImportForm):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1ImportForm):
            return True

        return self.to_dict() != other.to_dict()
